import logging

logger = logging.getLogger(__name__)
logger.warning(
    "y.utils.dank_mids.dank_w3 has been deprecated and will be removed soon. Please import and use `dank_web3` from `dank_mids` directly."
)
from dank_mids import dank_web3 as dank_w3



import logging
from itertools import chain
from typing import AsyncIterator, List, Optional

import a_sync
import dank_mids
import msgspec.json
import pony.orm
from a_sync.executor import _AsyncExecutorMixin
from evmspec import FilterTrace

from y._db.common import DiskCache, Filter, _clean_addresses, default_filter_threads
from y._db.decorators import a_sync_write_db_session
from y._db.entities import Chain, Trace, TraceCacheInfo, insert
from y._db.utils._ep import _get_get_block
from y.utils.middleware import BATCH_SIZE

logger = logging.getLogger(__name__)


@a_sync_write_db_session
def insert_trace(trace: FilterTrace) -> None:
    get_block = _get_get_block()
    kwargs = {
        "block": get_block(trace.blockNumber, sync=True),
        "hash": trace.transactionHash,
        "raw": msgspec.json.encode(trace),
    }
    for dct in [trace, *trace.values()]:
        if "from" in dct:
            kwargs["from_address"] = dct["from"]
        if "to" in dct:
            kwargs["to_address"] = dct["to"]
    insert(type=Trace, **kwargs)


class TraceCache(DiskCache[dict, TraceCacheInfo]):
    __slots__ = "from_addresses", "to_addresses"

    def __init__(self, from_addresses: List[str], to_addresses: List[str]):
        self.from_addresses = _clean_addresses(from_addresses)
        self.to_addresses = _clean_addresses(to_addresses)

    def load_metadata(
        self, chain: Chain, from_address: Optional[str], to_address: Optional[str]
    ) -> Optional[TraceCacheInfo]:
        return TraceCacheInfo.get(
            chain=chain, from_address=str(from_address), to_address=str(to_address)
        )

    def _is_cached_thru(self, from_block: int) -> int:
        from y._db.utils import utils as db

        chain = db.get_chain(sync=True)
        infos = [
            self.load_metadata(chain, from_address, to_address)
            or self.load_metadata(chain, from_address, None)
            or self.load_metadata(chain, None, to_address)
            for from_address in self.from_addresses
            for to_address in self.to_addresses
        ]
        if all(info and info.cached_from <= from_block for info in infos):
            return max(info.cached_thru for info in infos)
        return 0

    def _select(self, from_block: int, to_block: int) -> List[dict]:
        from y._db.utils import utils as db

        return [
            msgspec.json.decode(trace)
            for trace in pony.orm.select(
                trace.raw
                for trace in Trace
                if trace.block.chain == db.get_chain(sync=True)
                and (not self.from_addresses or trace.address in self.from_addresses)
                and (not self.to_addresses or trace.address in self.to_addresses)
                and trace.block.number >= from_block
                and trace.block.number <= to_block
            )
        ]

    def _set_metadata(self, from_block: int, done_thru: int) -> None:
        from y._db.utils import utils as db

        chain = db.get_chain(sync=True)
        should_commit = False
        if self.to_addresses and self.from_addresses:
            for from_address in self.from_addresses:
                for to_address in self.to_addresses:
                    if info := TraceCacheInfo.get(
                        chain=chain,
                        from_address=msgspec.json.encode([from_address]),
                        to_address=msgspec.json.encode([to_address]),
                    ):
                        if from_block < info.cached_from:
                            info.cached_from = from_block
                            should_commit = True
                        if done_thru > info.cached_thru:
                            info.cached_thru = done_thru
                            should_commit = True
                    else:
                        TraceCacheInfo(
                            chain=chain,
                            from_address=msgspec.json.encode([from_address]),
                            to_address=msgspec.json.encode([to_address]),
                        )
                        should_commit = True
        elif self.from_addresses:
            for from_address in self.from_addresses:
                if info := TraceCacheInfo.get(
                    chain=chain,
                    from_address=msgspec.json.encode([from_address]),
                    to_address=msgspec.json.encode([]),
                ):
                    if from_block < info.cached_from:
                        info.cached_from = from_block
                        should_commit = True
                    if done_thru > info.cached_thru:
                        info.cached_thru = done_thru
                        should_commit = True
                else:
                    TraceCacheInfo(
                        chain=chain,
                        from_address=msgspec.json.encode([from_address]),
                        to_address=msgspec.json.encode([]),
                    )
                    should_commit = True
        elif self.to_addresses:
            for to_address in self.to_addresses:
                if info := TraceCacheInfo.get(
                    chain=chain,
                    from_address=msgspec.json.encode([]),
                    to_address=msgspec.json.encode([to_address]),
                ):
                    if from_block < info.cached_from:
                        info.cached_from = from_block
                        should_commit = True
                    if done_thru > info.cached_thru:
                        info.cached_thru = done_thru
                        should_commit = True
                else:
                    TraceCacheInfo(
                        chain=chain,
                        from_address=msgspec.json.encode([]),
                        to_address=msgspec.json.encode([to_address]),
                    )
                    should_commit = True
        elif info := TraceCacheInfo.get(
            chain=chain,
            from_address=msgspec.json.encode([]),
            to_address=msgspec.json.encode([]),
        ):
            if from_block < info.cached_from:
                info.cached_from = from_block
                should_commit = True
            if done_thru > info.cached_thru:
                info.cached_thru = done_thru
                should_commit = True
        else:
            TraceCacheInfo(
                chain=chain,
                from_address=msgspec.json.encode([]),
                to_address=msgspec.json.encode([]),
            )
            should_commit = True

        if should_commit:
            pony.orm.commit()
            logger.debug(
                "cached %s %s thru %s",
                self.from_addresses,
                self.to_addresses,
                done_thru,
            )


class TraceFilter(Filter[dict, TraceCache]):
    insert_to_db = insert_trace
    __slots__ = "from_addresses", "to_addresses"

    def __init__(
        self,
        from_addresses: List[str],
        to_addresses: List[str],
        from_block: int,
        *,
        chunk_size: int = BATCH_SIZE,
        chunks_per_batch: Optional[int] = None,
        semaphore: Optional[dank_mids.BlockSemaphore] = None,
        executor: _AsyncExecutorMixin = default_filter_threads,
        is_reusable: bool = True,
        verbose: bool = False,
    ):
        self.from_addresses = from_addresses
        self.to_addresses = to_addresses
        super().__init__(
            from_block,
            chunk_size=chunk_size,
            chunks_per_batch=chunks_per_batch,
            semaphore=semaphore,
            executor=executor,
            is_reusable=is_reusable,
            verbose=verbose,
        )

    @property
    def cache(self) -> TraceCache:
        if self._cache is None:
            self._cache = TraceCache(self.from_addresses, self.to_addresses)
        return self._cache

    def traces(self, to_block: Optional[int]) -> AsyncIterator[dict]:
        return self._objects_thru(block=to_block)

    async def _fetch_range(self, from_block: int, to_block: int) -> List[dict]:
        try:
            return await dank_mids.web3.provider.make_request("TraceFilter", {})
        except NotImplementedError:
            tasks = a_sync.map(self._trace_block, range(from_block, to_block))
            results = {block: traces async for block, traces in tasks.map()}
            return list(chain(*[results[i] for i in range(from_block, to_block)]))

    async def _trace_block(self, block: int) -> List[dict]:
        return [
            trace
            for trace in await dank_mids.web3.provider.make_request("TraceBlock", block)
            if (not self.from_addresses or trace_is_from(self.from_addresses))
            and (not self.to_addresses or trace_is_to(self.to_addresses, trace))
        ]


trace_is_from = lambda addresses, trace: any(
    "from" in x and x["from"] in addresses for x in [trace, trace.values()]
)

trace_is_to = lambda addresses, trace: any(
    "to" in x and x["to"] in addresses for x in [trace, trace.values()]
)



import logging
from typing import Optional

import a_sync
from brownie import chain

from y import convert
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

POOL = "0x83E5f18Da720119fF363cF63417628eB0e9fd523"


def is_froyo(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return (
        chain.id == Network.Fantom
        and address == "0x4f85Bbf3B0265DCEd4Ec72ebD0358ccCf190F1B3"
    )


@a_sync.a_sync(default="sync")
async def get_price(
    token: AnyAddressType, block: Optional[Block] = None
) -> Optional[UsdPrice]:
    if not token == POOL:
        return None
    virtual_price = await raw_call(
        POOL, "get_virtual_price()", block=block, output="int", sync=False
    )
    return UsdPrice(virtual_price / 10**18)



import y
from tests.fixtures import async_test, optimism_only


@async_test
@optimism_only
async def test_wtbt():
    assert (
        await y.get_price(
            "0xdb4eA87fF83eB1c80b8976FC47731Da6a31D35e5",
            106500000,
            skip_cache=True,
            sync=False,
        )
        == 1.0009538615847542
    )


@async_test
@optimism_only
async def test_mseth():
    assert (
        await y.get_price(
            "0x1610e3c85dd44Af31eD7f33a63642012Dca0C5A5",
            106500000,
            skip_cache=True,
            sync=False,
        )
        == 1909.2991854198503
    )



import asyncio
import logging
from abc import abstractmethod, abstractproperty
from decimal import Decimal
from typing import Awaitable, List, Optional, Union

import a_sync
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from multicall import Call
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20, ContractBase
from y.contracts import Contract
from y.datatypes import Address, AddressOrContract, AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.utils import hasall
from y.utils.logging import get_price_logger
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


v1_pools = {
    Network.Mainnet: ["0x398eC7346DcD622eDc5ae82352F02bE94C62d119"],
}.get(chain.id, [])

v2_pools = {
    Network.Mainnet: [
        "0x7d2768dE32b0b80b7a3454c06BdAc94A69DDc7A9",  # aave
        "0x7937D4799803FbBe595ed57278Bc4cA21f3bFfCB",  # aave amm
        "0xcE744a9BAf573167B2CF138114BA32ed7De274Fa",  # umee
    ],
    Network.Polygon: [
        "0x8dFf5E27EA6b7AC08EbFdf9eB090F32ee9a30fcf",  # aave
    ],
    Network.Avalanche: [
        "0x70BbE4A294878a14CB3CDD9315f5EB490e346163",  # blizz
    ],
}.get(chain.id, [])

v3_pools = {
    Network.Mainnet: [
        "0x87870Bca3F3fD6335C3F4ce8392D69350B4fA4E2",  # aave v3
    ],
    Network.Optimism: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD",  # aave v3
    ],
    Network.Arbitrum: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD",  # aave v3
    ],
    Network.Harmony: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD",  # aave v3
    ],
    Network.Arbitrum: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD",  # aave v3
    ],
    Network.Fantom: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD",  # aave v3
    ],
    Network.Avalanche: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD",  # aave v3
    ],
    Network.Polygon: [
        "0x794a61358D6845594F94dc1DB02A252b5b4814aD",  # aave v3
    ],
}.get(chain.id, [])


class AaveMarketBase(ContractBase):
    """
    Base class for Aave markets.
    """

    def __contains__(self, token: object) -> bool:
        """
        Check if `token` is an aToken from this market.

        Args:
            atoken: The item to check.

        Returns:
            True if the token is an aToken from the market, False otherwise.
        """
        if not self.asynchronous:
            cls = self.__class__.__name__
            raise RuntimeError(
                f"'self.asynchronous' must be False to use {cls}.__contains__.\nYou may wish to use {cls}.is_atoken instead."
            )
        return convert.to_address(token) in self.atokens

    async def contains(self, token: object) -> bool:
        """
        Check if `token` is an aToken from this market.

        Args:
            token: The item to check.

        Returns:
            True if the token is an aToken from the market, False otherwise.
        """
        contains = await convert.to_address_async(token) in await self.__atokens__
        logger.debug("%s contains %s: %s", self, token, contains)
        return contains

    async def get_reserves(self) -> List[Address]:
        return await Call(self.address, [self._get_reserves_method])

    async def get_reserve_data(self, reserve: AnyAddressType) -> tuple:
        return await self.contract.getReserveData.coroutine(reserve)

    @abstractproperty
    async def atokens(self) -> Awaitable[List[ERC20]]:
        """
        Get the aTokens of the market.

        Returns:
            A list of aTokens as :class:`~ERC20` objects.
        """

    __atokens__: HiddenMethodDescriptor[Self, List[ERC20]]

    @abstractmethod
    async def underlying(self, atoken_address: AddressOrContract) -> ERC20:
        """
        Get the underlying asset of the given aToken address.

        Args:
            atoken_address: The address of the aToken.

        Returns:
            The underlying asset.
        """

    @abstractproperty
    def _get_reserves_method(self) -> str:
        """
        The method that must be called to get the reserves list.
        """


class AaveMarketV1(AaveMarketBase):
    @a_sync.aka.cached_property
    async def atokens(self) -> List[ERC20]:
        reserves_data = a_sync.map(self.get_reserve_data, self.get_reserves(sync=False))
        atokens = [
            ERC20(reserve_data["aTokenAddress"], asynchronous=self.asynchronous)
            async for _, reserve_data in reserves_data
        ]
        logger.info("loaded %s v1 atokens for %s", len(atokens), self)
        return atokens

    @a_sync.a_sync(ram_cache_maxsize=256)
    async def underlying(self, atoken_address: AddressOrContract) -> ERC20:
        underlying = await raw_call(
            atoken_address, "underlyingAssetAddress()", output="address", sync=False
        )
        return ERC20(underlying)

    _get_reserves_method = "getReserves()(address[])"


_V2_RESERVE_DATA_METHOD = "getReserveData(address)((uint256,uint128,uint128,uint128,uint128,uint128,uint40,address,address,address,address,uint8))"


class AaveMarketV2(AaveMarketBase):
    @a_sync.aka.cached_property
    async def atokens(self) -> List[ERC20]:
        reserves_data = a_sync.map(self.get_reserve_data, self.get_reserves(sync=False))
        try:
            atokens = [
                ERC20(reserve_data[7], asynchronous=self.asynchronous)
                async for _, reserve_data in reserves_data
            ]
            logger.info("loaded %s v2 atokens for %s", len(atokens), self)
            return atokens
        except (
            TypeError
        ) as e:  # TODO figure out what to do about non verified aave markets
            logger.exception(e)
            logger.warning("failed to load tokens for %s", self)
            return []

    async def get_reserve_data(self, reserve: AnyAddressType) -> tuple:
        return await Call(self.address, [_V2_RESERVE_DATA_METHOD, str(reserve)])

    @a_sync.a_sync(ram_cache_maxsize=256)
    async def underlying(self, atoken_address: AddressOrContract) -> ERC20:
        underlying = await raw_call(
            atoken_address, "UNDERLYING_ASSET_ADDRESS()", output="address", sync=False
        )
        logger.debug("underlying: %s", underlying)
        return ERC20(underlying, asynchronous=self.asynchronous)

    _get_reserves_method = "getReservesList()(address[])"


class AaveMarketV3(AaveMarketBase):
    @a_sync.aka.cached_property
    async def atokens(self) -> List[ERC20]:
        reserves_data = a_sync.map(self.get_reserve_data, self.get_reserves(sync=False))
        try:
            atokens = [
                ERC20(reserve_data[8], asynchronous=self.asynchronous)
                async for _, reserve_data in reserves_data
            ]
            logger.info("loaded %s v3 atokens for %s", len(atokens), self)
            return atokens
        except (
            TypeError
        ) as e:  # TODO figure out what to do about non verified aave markets
            logger.exception(e)
            logger.warning("failed to load tokens for %s", self)
            return []

    @a_sync.a_sync(ram_cache_maxsize=256)
    async def underlying(self, atoken_address: AddressOrContract) -> ERC20:
        underlying = await raw_call(
            atoken_address, "UNDERLYING_ASSET_ADDRESS()", output="address", sync=False
        )
        logger.debug("underlying: %s", underlying)
        return ERC20(underlying, asynchronous=self.asynchronous)

    _get_reserves_method = "getReservesList()(address[])"


AaveMarket = Union[AaveMarketV1, AaveMarketV2, AaveMarketV3]

_WRAPPED_V2_METHODS = "ATOKEN", "STATIC_ATOKEN_LM_REVISION", "staticToDynamicAmount"
_WRAPPED_V3_METHODS = "ATOKEN", "AAVE_POOL", "UNDERLYING"


class AaveRegistry(a_sync.ASyncGenericSingleton):
    def __init__(self, *, asynchronous: bool = False) -> None:
        self.asynchronous = asynchronous
        super().__init__()

    @a_sync.aka.cached_property
    async def pools(self) -> List[AaveMarket]:
        v1, v2, v3 = await asyncio.gather(
            self.__pools_v1__,
            self.__pools_v2__,
            self.__pools_v3__,
        )
        return v1 + v2 + v3

    __pools__: HiddenMethodDescriptor[Self, List[AaveMarket]]

    @a_sync.aka.cached_property
    async def pools_v1(self) -> List[AaveMarketV1]:
        pools = [
            AaveMarketV1(pool, asynchronous=self.asynchronous) for pool in v1_pools
        ]
        logger.debug("%s v1 pools %s", self, pools)
        return pools

    __pools_v1__: HiddenMethodDescriptor[Self, List[AaveMarketV1]]

    @a_sync.aka.cached_property
    async def pools_v2(self) -> List[AaveMarketV2]:
        pools = [
            AaveMarketV2(pool, asynchronous=self.asynchronous) for pool in v2_pools
        ]
        logger.debug("%s v2 pools %s", self, pools)
        return pools

    __pools_v2__: HiddenMethodDescriptor[Self, List[AaveMarketV2]]

    @a_sync.aka.cached_property
    async def pools_v3(self) -> List[AaveMarketV3]:
        pools = [
            AaveMarketV3(pool, asynchronous=self.asynchronous) for pool in v3_pools
        ]
        logger.debug("%s v3 pools %s", self, pools)
        return pools

    __pools_v3__: HiddenMethodDescriptor[Self, List[AaveMarketV3]]

    async def pool_for_atoken(
        self, atoken_address: AnyAddressType
    ) -> Optional[Union[AaveMarketV1, AaveMarketV2, AaveMarketV3]]:
        pools = await self.__pools__
        for pool in pools:
            if await pool.contains(atoken_address, sync=False):
                return pool

    def __contains__(self, __o: object) -> bool:
        if self.asynchronous:
            raise RuntimeError(
                f"'self.asynchronous' must be False to use AaveRegistry.__contains__.\nYou may wish to use AaveRegistry.is_atoken instead."
            )
        return any(__o in pool for pool in self.pools)

    @a_sync.a_sync(cache_type="memory")
    async def is_atoken(self, atoken_address: AnyAddressType) -> bool:
        logger = get_price_logger(atoken_address, block=None, extra="aave")
        is_atoken = any(
            await asyncio.gather(
                *[
                    pool.contains(atoken_address, sync=False)
                    for pool in await self.__pools__
                ]
            )
        )
        logger.debug("is_atoken: %s", is_atoken)
        return is_atoken

    async def is_wrapped_atoken_v2(self, atoken_address: AnyAddressType) -> bool:
        # NOTE: Not sure if this wrapped version is actually related to aave but this works for pricing purposes.
        contract = await Contract.coroutine(atoken_address, require_success=False)
        return contract.verified and hasall(contract, _WRAPPED_V2_METHODS)

    async def is_wrapped_atoken_v3(self, atoken_address: AnyAddressType) -> bool:
        # NOTE: Not sure if this wrapped version is actually related to aave but this works for pricing purposes.
        contract = await Contract.coroutine(atoken_address, require_success=False)
        return contract.verified and hasall(contract, _WRAPPED_V3_METHODS)

    @a_sync.a_sync(cache_type="memory")
    async def underlying(self, atoken_address: AddressOrContract) -> ERC20:
        pool: Union[AaveMarketV1, AaveMarketV2, AaveMarketV3] = (
            await self.pool_for_atoken(atoken_address, sync=False)
        )
        return await pool.underlying(atoken_address, sync=False)

    async def get_price(
        self,
        atoken_address: AddressOrContract,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> UsdPrice:
        underlying: ERC20 = await self.underlying(atoken_address, sync=False)
        return await underlying.price(block, skip_cache=skip_cache, sync=False)

    async def get_price_wrapped_v2(
        self,
        atoken_address: AddressOrContract,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        return await self._get_price_wrapped(
            atoken_address, "staticToDynamicAmount", block=block, skip_cache=skip_cache
        )

    async def get_price_wrapped_v3(
        self,
        atoken_address: AddressOrContract,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        return await self._get_price_wrapped(
            atoken_address, "convertToAssets", block=block, skip_cache=skip_cache
        )

    async def _get_price_wrapped(
        self,
        atoken_address: AddressOrContract,
        method: str,
        block: Optional[Block] = None,
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        contract, scale = await asyncio.gather(
            Contract.coroutine(atoken_address),
            ERC20(atoken_address, asynchronous=True).scale,
        )
        try:
            underlying, price_per_share = await asyncio.gather(
                contract.ATOKEN.coroutine(
                    block_identifier=block
                ),  # NOTE: We can probably cache this without breaking anything
                getattr(contract, method).coroutine(scale, block_identifier=block),
            )
        except ContractLogicError:
            return None
        price_per_share /= Decimal(scale)
        return price_per_share * Decimal(
            await ERC20(underlying, asynchronous=True).price(
                block, skip_cache=skip_cache
            )
        )


aave: AaveRegistry = AaveRegistry(asynchronous=True)



import asyncio
import logging
from decimal import Decimal
from typing import List, Optional

import a_sync
from multicall import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, WeiBalance
from y.contracts import Contract, has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.utils.cache import optional_async_diskcache

logger = logging.getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
@optional_async_diskcache
async def is_token_set(token: AnyAddressType) -> bool:
    return any(
        await asyncio.gather(
            has_methods(
                token, ("getComponents()(address[])", "naturalUnit()(uint)"), sync=False
            ),
            has_methods(
                token,
                (
                    "getComponents()(address[])",
                    "getModules()(address[])",
                    "getPositions()(address[])",
                ),
                sync=False,
            ),
        )
    )


@a_sync.a_sync(default="sync")
async def get_price(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    return await TokenSet(token, asynchronous=True).get_price(
        block=block, skip_cache=skip_cache
    )


class TokenSet(ERC20):
    def __init__(self, address: AnyAddressType, *, asynchronous: bool = False):
        super().__init__(address, asynchronous=asynchronous)
        self.get_total_component_real_units = Call(
            self.address, "getTotalComponentRealUnits(address)(int256)"
        )

    @a_sync.a_sync(ram_cache_maxsize=100)
    async def components(self, block: Optional[Block] = None) -> List[ERC20]:
        contract = await Contract.coroutine(self.address)
        components = await contract.getComponents.coroutine(block_identifier=block)
        return [
            ERC20(component, asynchronous=self.asynchronous) for component in components
        ]

    async def balances(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> List[WeiBalance]:
        contract, components = await asyncio.gather(
            Contract.coroutine(self.address),
            self.components(block=block, sync=False),
        )
        if hasattr(contract, "getUnits"):
            balances = await contract.getUnits.coroutine(block_identifier=block)
            logger.debug("getUnits: %s", balances)
        elif hasattr(contract, "getTotalComponentRealUnits"):
            _components = ((component.address,) for component in components)
            balances = await a_sync.map(
                self.get_total_component_real_units.coroutine,
                _components,
                block_id=block,
            ).values(pop=True)
            logger.debug("getTotalComponentRealUnits: %s", balances)
        balances = [
            WeiBalance(balance, component, block=block, skip_cache=skip_cache)
            for component, balance in zip(components, balances)
        ]
        return balances

    async def get_price(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> UsdPrice:
        total_supply = Decimal(
            await self.total_supply_readable(block=block, sync=False)
        )
        if total_supply == 0:
            logger.debug("total supply is 0, forcing price to $0")
            return UsdPrice(0)
        contract = await Contract.coroutine(self.address)
        if hasattr(contract, "getUnits"):
            balances: List[WeiBalance] = await self.balances(
                block=block, skip_cache=skip_cache, sync=False
            )
            values = await WeiBalance.value_usd.map(balances).values(pop=True)
            logger.debug("balances: %s  values: %s", balances, values)
            tvl = sum(values)
            price = UsdPrice(tvl / total_supply)
            logger.debug(
                "total supply: %s  tvl: %s  price: %s", total_supply, tvl, price
            )
            return price
        elif hasattr(contract, "getTotalComponentRealUnits"):
            balances_per_token: List[WeiBalance] = await self.balances(
                block=block, sync=False
            )
            price = UsdPrice(
                await WeiBalance.value_usd.sum(balances_per_token, sync=False)
            )
            logger.debug("balances per token: %s  price: %s", balances_per_token, price)
            return price
        raise NotImplementedError



import logging
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, Any
from typing import Optional as typing_Optional

import a_sync
from pony.orm import (
    Database,
    InterfaceError,
    Optional,
    PrimaryKey,
    Required,
    Set,
    TransactionIntegrityError,
    commit,
    composite_index,
    db_session,
)

from y._db.decorators import retry_locked, ydb_write_threads

db = Database()

# makes type checking work, see below for info:
# https://pypi.org/project/pony-stubs/
DbEntity = db.Entity


logger = logging.getLogger(__name__)


class _AsyncEntityMixin:
    # TODO: implement this
    pass
    """
    @classmethod
    @a_sync('async')
    def get(cls: E, *args, **kwargs) -> E:
        return super(DbEntity).get(*args, **kwargs)
    @classmethod
    @a_sync('async')
    def select(cls: E, *args, **kwargs) -> E:
        return super(DbEntity).select(*args, **kwargs)
    """


class Chain(DbEntity, _AsyncEntityMixin):
    id = PrimaryKey(int)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        blocks: Set["Block"]
        addresses: Set["Address"]
        log_cached: Set["LogCacheInfo"]
        trace_caches: Set["TraceCacheInfo"]

    blocks = Set("Block")
    addresses = Set("Address")
    log_cached = Set("LogCacheInfo")
    trace_caches = Set("TraceCacheInfo")


class Block(DbEntity, _AsyncEntityMixin):
    chain = Required(Chain, reverse="blocks")
    number = Required(int)
    PrimaryKey(chain, number)
    hash = Optional(int, lazy=True)
    timestamp = Optional(datetime, lazy=True)

    composite_index(chain, hash)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        prices: Set["Price"]
        contracts_deployed: Set["Contract"]
        logs: Set["Log"]
        traces: Set["Trace"]

    prices = Set("Price", reverse="block", cascade_delete=False)
    contracts_deployed = Set("Contract", reverse="deploy_block")
    logs = Set("Log", reverse="block", cascade_delete=False)
    traces = Set("Trace", reverse="block", cascade_delete=False)


class Address(DbEntity, _AsyncEntityMixin):
    chain = Required(Chain, lazy=True, reverse="addresses")
    address = Required(str, 42, lazy=True)
    PrimaryKey(chain, address)
    notes = Optional(str, lazy=True)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        contracts_deployed: Set["Contract"]

    contracts_deployed = Set("Contract", reverse="deployer")


class Contract(Address):
    deployer = Optional(
        Address, reverse="contracts_deployed", lazy=True, cascade_delete=False
    )
    deploy_block = Optional(
        Block, reverse="contracts_deployed", lazy=True, cascade_delete=False
    )


class Token(Contract):
    symbol = Optional(str, lazy=True)
    name = Optional(str, lazy=True)
    decimals = Optional(int, lazy=True)
    bucket = Optional(str, index=True, lazy=True)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        prices: Set["Price"]

    prices = Set("Price", reverse="token")


class Price(DbEntity):
    block = Required(Block, index=True, lazy=True)
    token = Required(Token, index=True, lazy=True)
    PrimaryKey(block, token)
    price = Required(Decimal, 38, 18)


class TraceCacheInfo(DbEntity):
    chain = Required(Chain, index=True)
    to_addresses = Required(bytes, index=True)
    from_addresses = Required(bytes, index=True)
    PrimaryKey(chain, to_addresses, from_addresses)
    cached_from = Required(int)
    cached_thru = Required(int)


class LogCacheInfo(DbEntity):
    chain = Required(Chain, index=True)
    address = Required(str, 42, index=True)
    topics = Required(bytes)
    PrimaryKey(chain, address, topics)
    cached_from = Required(int)
    cached_thru = Required(int)


class LogTopic(DbEntity):
    "Just makes the :ref:`Log` db smaller."
    dbid = PrimaryKey(int, size=64, auto=True)
    topic = Required(str, 64, unique=True, lazy=True)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        logs_as_topic0: Set["Log"]
        logs_as_topic1: Set["Log"]
        logs_as_topic2: Set["Log"]
        logs_as_topic3: Set["Log"]

    logs_as_topic0 = Set("Log", reverse="topic0")
    logs_as_topic1 = Set("Log", reverse="topic1")
    logs_as_topic2 = Set("Log", reverse="topic2")
    logs_as_topic3 = Set("Log", reverse="topic3")


class Hashes(DbEntity):
    "Just makes :ref:`Log` pk and indexes smaller."
    dbid = PrimaryKey(int, size=64, auto=True)
    hash = Required(str, 64, unique=True)

    if TYPE_CHECKING:
        # if we execute this code we get `TypeError: 'type' object is not subscriptable`
        logs_for_tx: Set["Log"]
        logs_for_addr: Set["Log"]

    logs_for_tx = Set("Log", reverse="tx")
    logs_for_addr = Set("Log", reverse="address")


class Log(DbEntity):
    block = Required(Block, index=True, lazy=True)
    tx = Required(Hashes, lazy=True)
    log_index = Required(int, size=16, lazy=True)
    PrimaryKey(block, tx, log_index)

    address = Required(Hashes, index=True, lazy=True)

    topic0 = Required(LogTopic, index=True, lazy=True)
    topic1 = Optional(LogTopic, index=True, lazy=True)
    topic2 = Optional(LogTopic, index=True, lazy=True)
    topic3 = Optional(LogTopic, index=True, lazy=True)
    composite_index(address, topic0)
    composite_index(topic0, topic1)
    composite_index(topic0, topic2)
    composite_index(topic0, topic3)
    composite_index(block, address, topic0)
    composite_index(block, topic0, topic1)
    composite_index(block, topic0, topic2)
    composite_index(block, topic0, topic3)
    composite_index(topic0, topic1, topic2, topic3)
    composite_index(block, topic0, topic1, topic2, topic3)
    raw = Required(bytes, lazy=True)


class Trace(DbEntity):
    id = PrimaryKey(int, auto=True)
    block = Required(Block, index=True, lazy=True)
    hash = Required(str, 64, index=True, lazy=True)
    from_address = Required(str, 42, index=True, lazy=True)
    to_address = Required(str, 42, index=True, lazy=True)
    raw = Required(bytes)


class BlockAtTimestamp(DbEntity):
    chainid = Required(int)
    timestamp = Required(datetime)
    PrimaryKey(chainid, timestamp)
    block = Required(int)


@a_sync.a_sync(executor=ydb_write_threads)
@db_session
@retry_locked
def insert(type: DbEntity, **kwargs: Any) -> typing_Optional[DbEntity]:
    try:
        while True:
            try:
                entity = type(**kwargs)
                commit()
                logger.debug("inserted %s to db", entity)
                return entity
            except InterfaceError as e:
                logger.debug("%s while inserting %s", e, type.__name__)
    except TransactionIntegrityError as e:
        constraint_errs = (
            "UNIQUE constraint failed",
            "duplicate key value violates unique constraint",
        )
        if any(err in (msg := str(e)) for err in constraint_errs):
            logger.debug("%s: %s %s", msg, type.__name__, kwargs)
        else:
            logger.debug(
                "%s %s when inserting %s",
                e.__class__.__name__,
                str(e),
                e,
                type.__name__,
            )
            raise



import errno
from os import mkdir, path

from y import ENVIRONMENT_VARIABLES as ENVS

SQLITE_DIR = f"{path.expanduser( '~' )}/.ypricemagic"
SQLITE_PATH = f"{SQLITE_DIR}/ypricemagic.sqlite"

if ENVS.DB_PROVIDER == "sqlite":
    try:
        mkdir(SQLITE_DIR)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    connection_settings = {
        "provider": str(ENVS.DB_PROVIDER),
        "filename": SQLITE_PATH,
        "create_db": True,
    }
else:
    connection_settings = {
        "provider": str(ENVS.DB_PROVIDER),
        "host": str(ENVS.DB_HOST),
        "user": str(ENVS.DB_USER),
        "password": str(ENVS.DB_PASSWORD),
        "database": str(ENVS.DB_DATABASE),
    }
    if ENVS.DB_PORT:
        connection_settings["port"] = int(ENVS.DB_PORT)



import asyncio
import logging
from contextlib import suppress
from decimal import Decimal
from functools import cached_property
from typing import Any, AsyncIterator, Dict, List, Optional, Tuple

import a_sync
import a_sync.exceptions
import brownie
import dank_mids
from a_sync.a_sync import HiddenMethodDescriptor
from brownie import chain
from brownie.network.event import _EventItem
from dank_mids.exceptions import Revert
from eth_typing import HexAddress
from multicall import Call
from typing_extensions import Self

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y._decorators import continue_on_revert, stuck_coro_debugger
from y.classes.common import ERC20, ContractBase, WeiBalance
from y.constants import STABLECOINS, WRAPPED_GAS_COIN, sushi, usdc, weth
from y.contracts import Contract, contract_creation_block_async
from y.datatypes import (
    Address,
    AddressOrContract,
    AnyAddressType,
    Block,
    Pool,
    UsdPrice,
)
from y.exceptions import (
    CantFindSwapPath,
    ContractNotVerified,
    NonStandardERC20,
    NotAUniswapV2Pool,
    TokenNotFound,
    call_reverted,
    continue_if_call_reverted,
    reraise_excs_with_extra_context,
)
from y.interfaces.uniswap.factoryv2 import UNIV2_FACTORY_ABI
from y.networks import Network
from y.prices import magic
from y.prices.dex.uniswap.v2_forks import (
    ROUTER_TO_FACTORY,
    ROUTER_TO_PROTOCOL,
    special_paths,
)
from y.utils.events import ProcessedEvents
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)

Path = List[AddressOrContract]
Reserves = Tuple[int, int, int]

factory_helper_address = {
    # put special case addresses here
}.get(chain.id, "0xE57Bfd650A7771E401d56d4b2CA22d9f8f51D3D9")

try:
    FACTORY_HELPER = Contract(factory_helper_address)
except ContractNotVerified:
    FACTORY_HELPER = None


class UniswapV2Pool(ERC20):
    """
    Represents a Uniswap V2 liquidity pool.
    """

    # default stored as class var to keep instance dicts smaller
    __types_assumed = True
    "True if we're assuming types based on normal univ2 abi, False if we checked via block explorer."

    __slots__ = ("get_reserves",)

    def __init__(
        self,
        address: AnyAddressType,
        token0: Optional[Address] = None,
        token1: Optional[Address] = None,
        deploy_block: Optional[int] = None,
        asynchronous: bool = False,
    ):
        super().__init__(address, asynchronous=asynchronous)
        self.get_reserves = Call(
            self.address, "getReserves()((uint112,uint112,uint32))"
        )
        if deploy_block:
            self._deploy_block = deploy_block
        if token0:
            self.token0 = token0
        if token1:
            self.token1 = token1

    @a_sync.aka.cached_property
    async def factory(self) -> Address:
        try:
            return await raw_call(
                self.address, "factory()", output="address", sync=False
            )
        except ValueError as e:
            if call_reverted(e):
                raise NotAUniswapV2Pool(self) from e
            # `is not a valid ETH address` means we got some kind of response from the chain.
            # but couldn't convert to address. If it happens to be a goofy but
            # verified uni fork, maybe we can get factory this way
            okay_errors = [
                "is not a valid ETH address",
                "invalid opcode",
                "invalid jump destination",
            ]
            if all(msg not in str(e) for msg in okay_errors):
                raise
            contract = await Contract.coroutine(self.address)
            try:
                return await contract.factory
            except AttributeError as exc:
                raise NotAUniswapV2Pool(self) from exc

    __factory__: HiddenMethodDescriptor[Self, Address]

    @a_sync.aka.property
    async def tokens(self) -> Tuple[ERC20, ERC20]:
        return await asyncio.gather(self.__token0__, self.__token1__)

    __tokens__: HiddenMethodDescriptor[Self, Tuple[ERC20, ERC20]]

    @a_sync.aka.cached_property
    async def token0(self) -> ERC20:
        try:
            if token0 := await Call(self.address, ["token0()(address)"]):
                return ERC20(token0, asynchronous=self.asynchronous)
        except ValueError as e:
            continue_if_call_reverted(e)
        raise NotAUniswapV2Pool(self)

    __token0__: HiddenMethodDescriptor[Self, ERC20]

    @a_sync.aka.cached_property
    async def token1(self) -> ERC20:
        try:
            if token1 := await Call(self.address, ["token1()(address)"]):
                return ERC20(token1, asynchronous=self.asynchronous)
        except ValueError as e:
            continue_if_call_reverted(e)
        raise NotAUniswapV2Pool(self)

    __token1__: HiddenMethodDescriptor[Self, ERC20]

    @a_sync.a_sync(ram_cache_ttl=ENVS.CACHE_TTL)
    @stuck_coro_debugger
    async def get_price(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Optional[UsdPrice]:
        tvl = await self.tvl(block=block, skip_cache=skip_cache, sync=False)
        if tvl is not None:
            return UsdPrice(
                tvl / Decimal(await self.total_supply_readable(block=block, sync=False))
            )

    @a_sync.a_sync(ram_cache_maxsize=None, ram_cache_ttl=ENVS.CACHE_TTL)
    async def get_token_out(self, token_in: Address) -> ERC20:
        if token_in == (token0 := await self.__token0__):
            # these return instantly since theyre already cached
            return await self.__token1__
        elif token_in == (token1 := await self.__token1__):
            # these return instantly since theyre already cached
            return await self.__token0__
        raise TokenNotFound(token_in, [token0, token1]) from None

    @stuck_coro_debugger
    async def reserves(
        self, *, block: Optional[Block] = None
    ) -> Optional[Tuple[WeiBalance, WeiBalance]]:
        try:
            reserves = await self.get_reserves.coroutine(block_id=block)
        except Exception as e:
            if not call_reverted(e):
                raise
            reserves = None

        if reserves is None and self.__types_assumed:
            try:
                await self._check_return_types()
            except AttributeError as e:
                raise NotAUniswapV2Pool(self) from e
            return await self.reserves(block=block, sync=False)

        if reserves is None and self._verified:
            # This shouldn't really run anymore, maybe delete
            contract = await Contract.coroutine(self.address)
            try:
                reserves = await contract.getReserves.coroutine(block_identifier=block)
                types = ",".join(
                    output["type"] for output in contract.getReserves.abi["outputs"]
                )
                logger.warning(f"abi for getReserves for {contract} is {types}")
            except Exception as e:
                if not call_reverted(e):
                    raise

        if reserves is None:
            return None

        # NOTE: using `__token0__` and `__token1__` is faster than `__tokens__` since they're already cached and return instantly
        #       it also creates 2 fewer tasks and 1 fewer future than `__tokens__` since there is no use of `asyncio.gather`.
        return (
            WeiBalance(reserves[0], await self.__token0__, block=block),
            WeiBalance(reserves[1], await self.__token1__, block=block),
        )

    @stuck_coro_debugger
    async def tvl(
        self, block: Optional[Block] = None, skip_cache: bool = ENVS.SKIP_CACHE
    ) -> Optional[Decimal]:
        # start these tasks now
        price_tasks: a_sync.TaskMapping[ERC20, UsdPrice]
        price_tasks = ERC20.price.map(
            # NOTE: using `__token0__` and `__token1__` is faster than `__tokens__` since they're already cached and return instantly
            #       it also creates 2 fewer tasks and 1 fewer future than `__tokens__` since there is no use of `asyncio.gather`.
            [await self.__token0__, await self.__token1__],
            block=block,
            return_None_on_failure=True,
            skip_cache=skip_cache,
        )

        reserves: Tuple[WeiBalance, WeiBalance]
        if (reserves := await self.reserves(block=block, sync=False)) is None:
            await price_tasks.close()
            return None

        prices = await price_tasks.values(pop=True)
        if vals := [
            Decimal(await reserve.__readable__) * Decimal(price)
            for reserve, price in zip(reserves, prices)
            if price is not None
        ]:
            if len(vals) == 1:
                vals *= 2
            if len(vals) == 2:
                logger.debug("reserves: %s", reserves)
                logger.debug("prices: %s", prices)
                logger.debug("vals: %s", vals)
                return sum(vals)
            else:
                raise Exception("how did we get here?") from None

    @stuck_coro_debugger
    @a_sync.a_sync(
        ram_cache_maxsize=100_000, ram_cache_ttl=60 * 60, semaphore=10_000
    )  # lets try a semaphore here
    async def check_liquidity(self, token: Address, block: Block) -> int:
        """
        Check the liquidity of a specific token in the pool at a given block.

        Args:
            token: The address of the token to check.
            block: The block number to query.

        Returns:
            The liquidity of the token in the pool.

        Raises:
            TokenNotFound: If the token is not one of the two tokens in the liquidity pool.
        """
        logger.debug("checking %s liquidity for %s at %s", self, token, block)
        if block and block < await self.deploy_block(sync=False):
            logger.debug("block %s is before %s deploy block", block, self)
            return 0
        if reserves := await self.reserves(block=block, sync=False):
            balance: WeiBalance
            for balance in reserves:
                if token == balance.token:
                    liquidity = balance.balance
                    logger.debug(
                        "%s liquidity for %s at %s is %s", self, token, block, liquidity
                    )
                    return liquidity
            raise TokenNotFound(token, reserves)
        return 0

    @stuck_coro_debugger
    async def is_uniswap_pool(self, block: Optional[Block] = None) -> bool:
        """
        Check if this contract is a valid liquidity pool for Uniswap V2 or one of its forks.

        Args:
            block (optional): The block number to query. Defaults to latest block.

        Returns:
            True if the contract is a valid Uniswap V2 pool, False otherwise.
        """
        try:
            return all(
                await asyncio.gather(
                    self.reserves(block=block, sync=False),
                    self.total_supply(block, sync=False),
                )
            )
        except (NotAUniswapV2Pool, ContractNotVerified):
            return False

    async def _check_return_types(self) -> None:
        if not self.__types_assumed:
            return
        try:
            contract = await Contract.coroutine(self.address)
            reserves_types = ",".join(
                output["type"] for output in contract.getReserves.abi["outputs"]
            )
            self._verified = True
            assert reserves_types.count(",") == 2, reserves_types
            self.get_reserves = Call(self.address, f"getReserves()(({reserves_types}))")
        except ContractNotVerified:
            self._verified = False
        self.__types_assumed = False


class PoolsFromEvents(ProcessedEvents[UniswapV2Pool]):
    PairCreated = "0x0d3648bd0f6ba80134a33ba9275ac585d9d315f0ad8355cddefde31afa28d0e9"
    __slots__ = "asynchronous", "label"

    def __init__(self, factory: AnyAddressType, label: str, asynchronous: bool = False):
        self.asynchronous = asynchronous
        self.label = label
        super().__init__(addresses=[factory], topics=[[self.PairCreated]])

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} label={self.label}>"

    def pools(self, to_block: Optional[int] = None) -> AsyncIterator[UniswapV2Pool]:
        return self._objects_thru(block=to_block)

    def _get_block_for_obj(self, obj: UniswapV2Pool) -> int:
        return obj._deploy_block

    def _process_event(self, event: _EventItem) -> UniswapV2Pool:
        pool = UniswapV2Pool(
            address=event["pair"],
            token0=event["token0"],
            token1=event["token1"],
            asynchronous=self.asynchronous,
        )
        # Do this here instead of in the init in case the user inited their own UniswapV2Pool object previoulsy, which is now the singleton
        pool._deploy_block = event.block_number
        return pool


def _log_factory_helper_failure(
    e: Exception, token_address, block, _ignore_pools
) -> str:
    stre = f"{e}"
    if "timeout" in stre:
        msg = "timeout"
    elif "out of gas" in stre:
        msg = "out of gas"
    elif isinstance(e, Revert):
        # TODO: debug me!
        msg = "reverted"
    else:
        raise e

    logger.debug(
        "helper %s for %s at block %s ignore_pools %s: %s",
        msg,
        token_address,
        block,
        _ignore_pools,
        e,
    )


_all_pools_semaphore = a_sync.Semaphore(
    10, name=f"{__name__}.UniswapRouterV2.all_pools_for"
)


class UniswapRouterV2(ContractBase):
    def __init__(
        self, router_address: AnyAddressType, *, asynchronous: bool = False
    ) -> None:
        super().__init__(router_address, asynchronous=asynchronous)

        self.label = ROUTER_TO_PROTOCOL[self.address]
        self.factory = ROUTER_TO_FACTORY[self.address]
        self.special_paths = special_paths(self.address)
        self.get_amounts_out = Call(
            self.address, "getAmountsOut(uint,address[])(uint[])"
        )

        # we need the factory contract object cached in brownie so we can decode logs properly
        if not ContractBase(self.factory, asynchronous=self.asynchronous)._is_cached:
            brownie.Contract.from_abi(
                "UniClone Factory [forced]", self.factory, UNIV2_FACTORY_ABI
            )

        self._events = PoolsFromEvents(
            self.factory, self.label, asynchronous=self.asynchronous
        )

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f"<UniswapV2Router {self.label} '{self.address}'>"

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def get_price(
        self,
        token_in: Address,
        block: Optional[Block] = None,
        token_out: Address = usdc.address,
        paired_against: Address = WRAPPED_GAS_COIN,
        skip_cache: bool = ENVS.SKIP_CACHE,
        ignore_pools: Tuple[Pool, ...] = (),
    ) -> Optional[UsdPrice]:
        """
        Calculate a price based on Uniswap Router quote for selling one `token_in`.
        Always uses intermediate WETH pair if `[token_in,weth,token_out]` swap path available.
        """

        token_in, token_out, path = str(token_in), str(token_out), None

        if chain.id == Network.BinanceSmartChain and token_out == usdc.address:
            busd = await Contract.coroutine(
                "0xe9e7CEA3DedcA5984780Bafc599bD69ADd087D56"
            )
            token_out = busd.address

        if token_in in STABLECOINS:
            return 1

        try:
            amount_in = await ERC20(token_in, asynchronous=True).scale
        except NonStandardERC20:
            return None

        if token_in in [weth.address, WRAPPED_GAS_COIN] and token_out in STABLECOINS:
            path = [token_in, token_out]

        elif str(token_out) in STABLECOINS:
            with suppress(CantFindSwapPath):
                path = await self.get_path_to_stables(
                    token_in, block, _ignore_pools=ignore_pools, sync=False
                )
                logger.debug("smrt")

        # If we can't find a good path to stables, we might still be able to determine price from price of paired token
        if path is None and (
            deepest_pool := await self.deepest_pool(
                token_in, block, _ignore_pools=ignore_pools, sync=False
            )
        ):
            logger.debug("deepest pool: %s", deepest_pool)
            paired_with = await deepest_pool.get_token_out(token_in, sync=False)
            path = [token_in, paired_with]
            quote, out_scale = await asyncio.gather(
                self.get_quote(amount_in, path, block=block, sync=False),
                ERC20(path[-1], asynchronous=True).scale,
            )
            logger.debug("quote: %s", quote)
            if quote is not None:
                amount_out = Decimal(quote[-1]) / out_scale
                fees = Decimal(0.997) ** (len(path) - 1)
                amount_out /= fees
                paired_with_price = await magic.get_price(
                    paired_with,
                    block,
                    fail_to_None=True,
                    skip_cache=skip_cache,
                    ignore_pools=(*ignore_pools, deepest_pool),
                    sync=False,
                )

                if paired_with_price:
                    return amount_out * Decimal(paired_with_price)

        # If we still don't have a workable path, try this smol brain method
        if path is None:
            path = self._smol_brain_path_selector(token_in, token_out, paired_against)
            # NOTE: does this ever run anymore? can we take it out?
            logger.warning("using smol brain path selector")

        fees = 0.997 ** (len(path) - 1)
        logger.debug("router: %s     path: %s", self.label, path)
        quote, out_scale = await asyncio.gather(
            self.get_quote(amount_in, path, block=block, sync=False),
            ERC20(path[-1], asynchronous=True).scale,
        )
        if quote is not None:
            amount_out = quote[-1] / out_scale
            return UsdPrice(amount_out / fees)

    @continue_on_revert
    @stuck_coro_debugger
    async def get_quote(
        self, amount_in: int, path: Path, block: Optional[Block] = None
    ) -> Tuple[int, int]:
        if not self._is_cached:
            return await self.get_amounts_out.coroutine(
                (amount_in, path), block_id=block
            )
        try:
            return await self.contract.getAmountsOut.coroutine(
                amount_in, path, block_identifier=block
            )
        # TODO figure out how to best handle uni forks with slight modifications.
        # Sometimes the below "else" code will not work with modified methods. Brownie works for now.
        except Exception as e:
            strings = [
                "INSUFFICIENT_INPUT_AMOUNT",
                "INSUFFICIENT_LIQUIDITY",
                "INSUFFICIENT_OUT_LIQUIDITY",
                "Sequence has incorrect length",
            ]
            if not call_reverted(e) and all(s not in str(e) for s in strings):
                raise

    @a_sync.aka.cached_property
    @stuck_coro_debugger
    async def pools(self) -> List[UniswapV2Pool]:
        logger.info(
            "Fetching pools for %s on %s. If this is your first time using ypricemagic, this can take a while. Please wait patiently...",
            self.label,
            Network.printable(),
        )
        pools = [
            pool
            async for pool in self._events.pools(
                to_block=await dank_mids.eth.block_number
            )
        ]
        all_pairs_len = await raw_call(
            self.factory, "allPairsLength()", output="int", sync=False
        )
        if len(pools) > all_pairs_len:
            raise NotImplementedError("this shouldnt happen again")
        elif to_get := all_pairs_len - len(pools):  # <
            logger.debug(
                "Oh no! Looks like your node can't look back that far. Checking for the missing %s pools...",
                to_get,
            )
            factory = await Contract.coroutine(self.factory)
            for i, pool_address in enumerate(await factory.allPairs.map(range(to_get))):
                pool = UniswapV2Pool(
                    address=pool_address, asynchronous=self.asynchronous
                )
                pools.insert(i, pool)
            logger.debug("Done fetching %s missing pools on %s", to_get, self.label)
        token0s = UniswapV2Pool.token0.map(pools).values(pop=True)
        token1s = UniswapV2Pool.token1.map(pools).values(pop=True)
        tokens = set(await token0s + await token1s)
        logger.info(
            "Loaded %s pools supporting %s tokens on %s",
            len(pools),
            len(tokens),
            self.label,
        )
        return pools

    __pools__: HiddenMethodDescriptor[Self, List[UniswapV2Pool]]

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=None, semaphore=_all_pools_semaphore)
    async def all_pools_for(self, token_in: Address) -> Dict[UniswapV2Pool, Address]:
        pool_to_token_out = {}
        for i, pool in enumerate(await self.__pools__):
            # these will return immediately since the pools are already loaded by this point
            if token_in == await pool.__token0__:
                pool_to_token_out[pool] = await pool.__token1__
            elif token_in == await pool.__token1__:
                pool_to_token_out[pool] = await pool.__token0__
            if not i % 10_000:
                await asyncio.sleep(0)
        return pool_to_token_out

    @stuck_coro_debugger
    async def get_pools_for(
        self, token_in: Address, block: Optional[Block] = None
    ) -> Dict[UniswapV2Pool, Address]:
        if not self._supports_uniswap_helper:
            return await self.all_pools_for(token_in, sync=False)
        try:
            pools: List[HexAddress] = await FACTORY_HELPER.getPairsFor.coroutine(
                self.factory, token_in, block_identifier=block
            )
        except Exception as e:
            if call_reverted(e) or "out of gas" in str(e) or "timeout" in str(e):
                return await self.all_pools_for(token_in, sync=False)
            raise

        pool_to_token_out = {}
        for p in pools:
            pool = UniswapV2Pool(p, asynchronous=self.asynchronous)
            # these will return immediately since the pools are already loaded by this point
            if token_in == await pool.__token0__:
                pool_to_token_out[pool] = await pool.__token1__
            elif token_in == await pool.__token1__:
                pool_to_token_out[pool] = await pool.__token0__
        return pool_to_token_out

    @stuck_coro_debugger
    async def pools_for_token(
        self,
        token_address: Address,
        block: Optional[Block] = None,
        _ignore_pools: Tuple[UniswapV2Pool, ...] = (),
    ) -> AsyncIterator[UniswapV2Pool]:
        pools: Dict[UniswapV2Pool, Address]

        if (
            chain.id == Network.Mainnet
            and token_address == WRAPPED_GAS_COIN
            and self.label == "uniswap v2"
        ):
            # This will run out of gas if we use the helper so we bypass it with a known liquid pool
            usdc_pool = UniswapV2Pool(
                "0xB4e16d0168e52d35CaCD2c6185b44281Ec28C9Dc", asynchronous=True
            )
            pools = {usdc_pool: "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"}
        else:
            pools = await self.get_pools_for(token_address, block=block, sync=False)

        for pool in _ignore_pools:
            pools.pop(pool, None)

        if not pools:
            return

        elif block is None:
            for pool in pools:
                yield pool
            return

        async for pool, deploy_block in ERC20.deploy_block.map(
            when_no_history_return_0=True
        ).map(pools):
            if deploy_block <= block:
                yield pool

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def deepest_pool(
        self,
        token_address: AnyAddressType,
        block: Optional[Block] = None,
        _ignore_pools: Tuple[UniswapV2Pool, ...] = (),
    ) -> Optional[UniswapV2Pool]:
        """returns the deepest pool for `token_address` at `block`, excluding pools in `_ignore_pools`"""
        token_address = await convert.to_address_async(token_address)
        if token_address == WRAPPED_GAS_COIN or token_address in STABLECOINS:
            return await self.deepest_stable_pool(token_address, block, sync=False)
        if self._supports_uniswap_helper and (
            block is None
            or block >= await contract_creation_block_async(FACTORY_HELPER)
        ):
            try:
                deepest_pool, deepest_pool_depth = await self.deepest_pool_for(
                    token_address, block, ignore_pools=_ignore_pools
                )
                return (
                    None
                    if deepest_pool == brownie.ZERO_ADDRESS
                    else UniswapV2Pool(deepest_pool, asynchronous=self.asynchronous)
                )
            except (Revert, ValueError) as e:
                msg = _log_factory_helper_failure(
                    e, token_address, block, _ignore_pools
                )

        pools = self.pools_for_token(token_address, block, _ignore_pools=_ignore_pools)
        logger.debug(
            "checking %s liquidity at block %s for pools %s",
            token_address,
            block,
            pools,
        )

        deepest_pool: UniswapV2Pool
        async for deepest_pool in (
            UniswapV2Pool.check_liquidity.map(pools, token=token_address, block=block)
            .keys(pop=True)
            .aiterbyvalues(reverse=True)
        ):
            return deepest_pool

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def deepest_stable_pool(
        self,
        token_address: AnyAddressType,
        block: Optional[Block] = None,
        _ignore_pools: Tuple[UniswapV2Pool, ...] = (),
    ) -> Optional[UniswapV2Pool]:
        """returns the deepest pool for `token_address` at `block` which has `token_address` paired with a stablecoin, excluding pools in `_ignore_pools`"""
        token_out_tasks: a_sync.TaskMapping[UniswapV2Pool, ERC20]
        deepest_stable_pool: UniswapV2Pool
        token_address = await convert.to_address_async(token_address)
        pools = self.pools_for_token(
            token_address, block=block, _ignore_pools=_ignore_pools
        )
        token_out_tasks = UniswapV2Pool.get_token_out.map(token_in=token_address)
        if stable_pools := [
            pool
            async for pool, paired_with in token_out_tasks.map(pools)
            if paired_with in STABLECOINS
        ]:
            del token_out_tasks

            if self._supports_uniswap_helper and (
                block is None
                or block
                >= await contract_creation_block_async(
                    FACTORY_HELPER, when_no_history_return_0=True
                )
            ):
                (
                    deepest_stable_pool,
                    deepest_stable_pool_balance,
                ) = await FACTORY_HELPER.deepestPoolForFrom.coroutine(
                    token_address, stable_pools, block_identifier=block
                )
                return (
                    None
                    if deepest_stable_pool == brownie.ZERO_ADDRESS
                    else UniswapV2Pool(
                        deepest_stable_pool, asynchronous=self.asynchronous
                    )
                )

            async for deepest_stable_pool, depth in (
                UniswapV2Pool.check_liquidity.map(
                    stable_pools, token=token_address, block=block
                )
                .items(pop=True)
                .aiterbyvalues(reverse=True)
            ):
                return deepest_stable_pool

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=500)
    async def get_path_to_stables(
        self,
        token: AnyAddressType,
        block: Optional[Block] = None,
        _loop_count: int = 0,
        _ignore_pools: Tuple[UniswapV2Pool, ...] = (),
    ) -> Path:
        if _loop_count > 10:
            raise CantFindSwapPath
        token_address = await convert.to_address_async(token)
        path = [token_address]
        deepest_pool = await self.deepest_pool(
            token_address, block, _ignore_pools, sync=False
        )
        if deepest_pool:
            paired_with = await deepest_pool.get_token_out(token_address, sync=False)
            from y.prices.utils.buckets import check_bucket

            if await check_bucket(paired_with, sync=False) and _loop_count == 0:
                # let's just use the other token to get the price
                return None
            deepest_stable_pool = await self.deepest_stable_pool(
                token_address, block, _ignore_pools=_ignore_pools, sync=False
            )
            if deepest_stable_pool and deepest_pool == deepest_stable_pool:
                path.append(
                    await deepest_stable_pool.get_token_out(token_address, sync=False)
                )
                return path

            if path == [token_address]:
                with suppress(CantFindSwapPath):
                    path.extend(
                        await self.get_path_to_stables(
                            paired_with,
                            block=block,
                            _loop_count=_loop_count + 1,
                            _ignore_pools=tuple(list(_ignore_pools) + [deepest_pool]),
                            sync=False,
                        )
                    )

        if path == [token_address]:
            raise CantFindSwapPath(
                f"Unable to find swap path for {token_address} on {Network.printable()}"
            )

        return path

    @stuck_coro_debugger
    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60 * 60)
    async def check_liquidity(
        self, token: Address, block: Block, ignore_pools=[]
    ) -> int:
        logger.debug("checking %s liquidity for %s at %s", self, token, block)
        if block and block < await contract_creation_block_async(self.factory):
            logger.debug("block %s is before %s deploy block")
            return 0
        if self._supports_uniswap_helper and (
            block is None
            or block >= await contract_creation_block_async(FACTORY_HELPER)
        ):
            try:
                deepest_pool, liquidity = await self.deepest_pool_for(
                    token, block, ignore_pools=ignore_pools
                )
                logger.debug(
                    "%s liquidity for %s at %s is %s", self, token, block, liquidity
                )
                return liquidity
            except (Revert, ValueError) as e:
                _log_factory_helper_failure(e, token, block, ignore_pools)

        pools = self.pools_for_token(token, block=block, _ignore_pools=ignore_pools)
        try:
            liquidity = await UniswapV2Pool.check_liquidity.max(
                pools, token=token, block=block, sync=False
            )
        except a_sync.exceptions.EmptySequenceError:
            liquidity = 0
        logger.debug("%s liquidity for %s at %s is %s", self, token, block, liquidity)
        return liquidity

    @a_sync.a_sync(ram_cache_maxsize=100_000, ram_cache_ttl=60 * 60)
    @stuck_coro_debugger
    async def deepest_pool_for(
        self, token: Address, block: Block = None, *, ignore_pools=[]
    ) -> Tuple[Address, int]:
        # sourcery skip: default-mutable-arg
        with reraise_excs_with_extra_context(self, token, block, ignore_pools):
            deepest = await FACTORY_HELPER.deepestPoolFor.coroutine(
                self.factory, token, ignore_pools, block_identifier=block
            )
            logger.debug(
                "got deepest pool for %s at %s: %s from helper", token, block, deepest
            )
            return deepest

    @cached_property
    def _supports_uniswap_helper(self) -> bool:
        """returns True if our uniswap helper contract is supported, False if not"""
        return (
            chain.id != Network.Mainnet
            and FACTORY_HELPER
            and self.label
            and self.label not in {"zipswap"}
        )

    def _smol_brain_path_selector(
        self,
        token_in: AddressOrContract,
        token_out: AddressOrContract,
        paired_against: AddressOrContract,
    ) -> Path:
        # sourcery skip: assign-if-exp, lift-return-into-if, merge-duplicate-blocks, merge-else-if-into-elif, remove-redundant-if, remove-unnecessary-cast
        """Chooses swap path to use for quote"""
        # NOTE: can we just delete this now? probably, must test
        token_in, token_out, paired_against = (
            str(token_in),
            str(token_out),
            str(paired_against),
        )

        if str(paired_against) in STABLECOINS and str(token_out) in STABLECOINS:
            path = [token_in, paired_against]
        elif weth in (token_in, token_out):
            path = [token_in, token_out]
        elif sushi and paired_against == sushi and token_out != sushi:
            path = [token_in, sushi, weth, token_out]
        elif str(token_in) in self.special_paths and str(token_out) in STABLECOINS:
            path = self.special_paths[str(token_in)]

        elif chain.id == Network.BinanceSmartChain:
            from y.constants import cake, wbnb

            if WRAPPED_GAS_COIN in (token_in, token_out):
                path = [token_in, token_out]
            elif cake.address in (token_in, token_out):
                path = [token_in, token_out]
            else:
                path = [token_in, wbnb.address, token_out]
        else:
            if WRAPPED_GAS_COIN in (token_in, token_out):
                path = [token_in, token_out]
            else:
                path = [token_in, WRAPPED_GAS_COIN, token_out]

        return path



from typing import final

import evmspec


@final
class Log(evmspec.log.Log, frozen=True, kw_only=True, array_like=True):
    """
    It works just like a :class:`~evmspec.log.Log` but it encodes to a tuple instead of a dict to save space since keys are known.
    """



import asyncio
import logging
from typing import Optional

import a_sync

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20
from y.contracts import Contract, has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.prices import magic

logger = logging.getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
async def is_mstable_feeder_pool(address: AnyAddressType) -> bool:
    return await has_methods(
        address, ("getPrice()((uint,uint))", "mAsset()(address)"), sync=False
    )


@a_sync.a_sync(default="sync")
async def get_price(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    address = await convert.to_address_async(token)
    contract = await Contract.coroutine(address)
    ratio, masset, scale = await asyncio.gather(
        contract.getPrice.coroutine(block_identifier=block),
        contract.mAsset.coroutine(block_identifier=block),
        ERC20(address, asynchronous=True).scale,
    )
    ratio = ratio[0] / scale
    underlying_price = await magic.get_price(
        masset, block, skip_cache=skip_cache, sync=False
    )
    return UsdPrice(underlying_price * ratio)



from functools import lru_cache
from typing import Optional, Set

import a_sync
import brownie.convert
import checksum_dict._key
import eth_event.main
import eth_utils
import evmspec.data
import multicall.call
import web3._utils as web3_utils
import web3.main as web3_main
import web3.middleware as web3_middleware
from brownie.convert.datatypes import HexBytes
from checksum_dict import to_checksum_address
from eth_typing import ChecksumAddress, HexAddress

from y import ENVIRONMENT_VARIABLES as ENVS
from y.datatypes import AnyAddressType


@lru_cache(maxsize=ENVS.CHECKSUM_CACHE_MAXSIZE)
def checksum(address: AnyAddressType) -> ChecksumAddress:
    try:
        return to_checksum_address(address)
    except ValueError:
        raise ValueError(f"'{address}' is not a valid ETH address") from None


def to_address(address: AnyAddressType) -> ChecksumAddress:
    address = __normalize_input_to_string(address)
    if checksummed := __get_checksum_from_cache(address):
        return checksummed
    checksummed = checksum(address)
    __cache_if_is_checksummed(address, checksummed)
    return checksummed


_checksum_thread = a_sync.ThreadPoolExecutor(1)


async def to_address_async(address: AnyAddressType) -> ChecksumAddress:
    address = __normalize_input_to_string(address)
    if checksummed := __get_checksum_from_cache(address):
        return checksummed
    checksummed = await _checksum_thread.run(checksum, address)
    __cache_if_is_checksummed(address, checksummed)
    return checksummed


_is_checksummed: Set[HexAddress] = set()
_is_not_checksummed: Set[HexAddress] = set()


def __get_checksum_from_cache(address: AnyAddressType) -> Optional[ChecksumAddress]:
    if address in _is_checksummed:
        return address
    elif address in _is_not_checksummed:
        # The checksum value is already in the lru-cache, there
        # is no reason to dispatch this function call to a thread.
        return checksum(address)


def __cache_if_is_checksummed(
    address: HexAddress, checksummed: ChecksumAddress
) -> None:
    if address == checksummed:
        _is_checksummed.add(address)
    else:
        _is_not_checksummed.add(address)


def __normalize_input_to_string(address: AnyAddressType) -> HexAddress:
    address_type = type(address)
    if address_type is str:
        return address
    elif issubclass(address_type, bytes):
        return (
            address.hex()
            if address_type.__name__ == "HexBytes"
            else HexBytes(address).hex()
        )
    elif address_type is int:
        return _int_to_address(address)
    else:
        return str(address)


def _int_to_address(int_address: int) -> HexAddress:
    hex_value = HexBytes(int_address).hex()[2:]
    padding = "0" * (40 - len(hex_value))
    return f"0x{padding}{hex_value}"


def _monkey_patch_dependencies():
    """
    Monkey patch dependency checksums with faster versions.
    """
    # this monkey patches brownie's convert.to_address with our lru_cache
    brownie.convert.to_address = to_address

    # this monkey patches checksum_dict's checksumming with our lru_cache
    checksum_dict._key.to_checksum_address = to_address

    # this monkey patches eth_event's address checksumming with our lru_cache
    eth_event.main.to_checksum_address = to_address

    # this monkey patches brownie's EthAddress class with faster execution
    eth_utils.to_checksum_address = to_address

    # this monkey patches something else I don't remember now with faster execution
    eth_utils.address.to_checksum_address = to_address

    # this monkey patches evmspec's Address decode hook with our lru cache
    evmspec.data.to_checksum_address = to_address

    # this monkey patches multicall.Call.target checksumming with our lru cache
    multicall.call.to_checksum_address = to_address

    # this monkey patches all checksums in web3py with faster execution
    web3_main.to_checksum_address = to_address
    web3_utils.ens.to_checksum_address = to_address
    web3_utils.method_formatters.to_checksum_address = to_address
    web3_utils.normalizers.to_checksum_address = to_address
    web3_middleware.signing.to_checksum_address = to_address

    try:
        import web3.utils.address as web3_address

        web3_address.to_checksum_address = to_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass

    try:
        import ens.ens

        ens.ens.to_checksum_address = to_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass

    try:
        import ens.async_ens

        ens.async_ens.to_checksum_address = to_address
    except ModuleNotFoundError:
        # youre on an older web3py, no monkey patch for you
        pass



import logging
import time
from functools import lru_cache, wraps
from typing import Callable, Iterable, TypeVar
from typing_extensions import ParamSpec

import a_sync
from a_sync.a_sync import ASyncFunction
from brownie import chain
from pony.orm import (
    CommitException,
    OperationalError,
    TransactionError,
    UnexpectedError,
    commit,
    db_session,
)

_T = TypeVar("_T")
_P = ParamSpec("_P")

logger = logging.getLogger(__name__)


ydb_read_threads = a_sync.PruningThreadPoolExecutor(12)
ydb_write_threads = a_sync.PruningThreadPoolExecutor(12)


def retry_locked(callable: Callable[_P, _T]) -> Callable[_P, _T]:
    @wraps(callable)
    def retry_locked_wrap(*args: _P.args, **kwargs: _P.kwargs) -> _T:
        sleep = 0.05
        while True:
            try:
                retval = callable(*args, **kwargs)
                commit()
                return retval
            except (CommitException, OperationalError, UnexpectedError) as e:
                log = logger.warning if sleep > 1 else logger.debug
                log("%s.%s got exc %s", callable.__module__, callable.__name__, e)
                if "database is locked" not in str(e):
                    raise
                time.sleep(sleep)
                sleep *= 1.5
            except TransactionError as e:
                logger.debug(
                    "%s.%s got exc %s", callable.__module__, callable.__name__, e
                )
                if (
                    "An attempt to mix objects belonging to different transactions"
                    not in str(e)
                ):
                    raise

    return retry_locked_wrap


a_sync_read_db_session: Callable[[Callable[_P, _T]], ASyncFunction[_P, _T]] = (
    lambda fn: a_sync.a_sync(default="async", executor=ydb_read_threads)(
        retry_locked(db_session(retry_locked(fn)))
    )
)

a_sync_write_db_session: Callable[[Callable[_P, _T]], ASyncFunction[_P, _T]] = (
    lambda fn: a_sync.a_sync(default="async", executor=ydb_write_threads)(
        retry_locked(db_session(retry_locked(fn)))
    )
)

a_sync_write_db_session_cached: Callable[[Callable[_P, _T]], ASyncFunction[_P, _T]] = (
    lambda fn: a_sync.a_sync(
        default="async", executor=ydb_write_threads, ram_cache_maxsize=None
    )(retry_locked(lru_cache(maxsize=None)(db_session(retry_locked(fn)))))
)


_result_count_logger = logging.getLogger(f"{__name__}.result_count")


def log_result_count(
    name: str, arg_names: Iterable[str] = []
) -> Callable[[Callable[_P, _T]], Callable[_P, _T]]:
    def result_count_deco(fn: Callable[_P, _T]) -> Callable[_P, _T]:
        @wraps(fn)
        def result_count_wrap(*args: _P.args, **kwargs: _P.kwargs) -> _T:
            results = fn(*args, **kwargs)
            if _result_count_logger.isEnabledFor(logging.DEBUG):
                arg_values = " ".join(
                    f"{k} {v}" for k, v in [("chain", chain.id), *zip(arg_names, args)]
                )
                _result_count_logger.debug(
                    "loaded %s %s for %s", len(results), name, arg_values
                )
            return results

        return result_count_wrap

    return result_count_deco



import asyncio
import logging
from typing import Optional

import a_sync
from brownie import chain

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20
from y.constants import weth
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.networks import Network
from y.prices import magic
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


def is_creth(token: AnyAddressType) -> bool:
    address = convert.to_address(token)
    return (
        chain.id == Network.Mainnet
        and address == "0xcBc1065255cBc3aB41a6868c22d1f1C573AB89fd"
    )


@a_sync.a_sync(default="sync")
async def get_price_creth(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    address = await convert.to_address_async(token)
    total_balance, total_supply, weth_price = await asyncio.gather(
        raw_call(address, "accumulated()", output="int", block=block, sync=False),
        ERC20(address, asynchronous=True).total_supply(block),
        magic.get_price(weth, block, skip_cache=skip_cache, sync=False),
    )
    per_share = total_balance / total_supply
    return per_share * weth_price



import asyncio
import logging
from typing import Optional

import a_sync

import y.prices.magic
from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20
from y.contracts import has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice
from y.utils.cache import optional_async_diskcache
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
@optional_async_diskcache
async def is_gelato_pool(token_address: AnyAddressType) -> bool:
    return await has_methods(
        token_address, ("gelatoBalance0()(uint)", "gelatoBalance1()(uint)"), sync=False
    )


@a_sync.a_sync(default="sync")
async def get_price(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    address = await convert.to_address_async(token)

    token0, token1 = await asyncio.gather(
        raw_call(address, "token0()", block=block, output="address", sync=False),
        raw_call(address, "token1()", block=block, output="address", sync=False),
    )

    (
        balance0,
        balance1,
        scale0,
        scale1,
        price0,
        price1,
        total_supply,
    ) = await asyncio.gather(
        raw_call(address, "gelatoBalance0()", block=block, output="int", sync=False),
        raw_call(address, "gelatoBalance1()", block=block, output="int", sync=False),
        ERC20(token0, asynchronous=True).scale,
        ERC20(token1, asynchronous=True).scale,
        y.prices.magic.get_price(token0, block, skip_cache=skip_cache, sync=False),
        y.prices.magic.get_price(token1, block, skip_cache=skip_cache, sync=False),
        ERC20(address, asynchronous=True).total_supply_readable(block, sync=False),
    )

    balance0 /= scale0
    balance1 /= scale1
    totalVal = balance0 * price0 + balance1 * price1
    return UsdPrice(totalVal / total_supply)



import abc
import asyncio
import logging
from functools import lru_cache
from typing import (
    Any,
    AsyncIterator,
    Awaitable,
    Callable,
    Container,
    Generic,
    List,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
)

import a_sync
import dank_mids
import eth_retry
from a_sync.executor import _AsyncExecutorMixin
from async_property import async_property
from brownie import ZERO_ADDRESS
from evmspec.data import Address, HexBytes32
from hexbytes import HexBytes
from pony.orm import OptimisticCheckError, TransactionIntegrityError, db_session
from web3.datastructures import AttributeDict
from web3.middleware.filter import block_ranges

from y import convert
from y._db.decorators import retry_locked
from y._db.exceptions import CacheNotPopulatedError
from y._decorators import stuck_coro_debugger
from y.exceptions import reraise_excs_with_extra_context
from y.utils.middleware import BATCH_SIZE

T = TypeVar("T")
S = TypeVar("S")
M = TypeVar("M")

logger = logging.getLogger(__name__)
default_filter_threads = a_sync.PruningThreadPoolExecutor(4)
"""
The thread pool executor used for all :class:`Filter` objects without one provided, with a maximum of 4 threads.
"""


def enc_hook(obj: Any) -> bytes:
    """
    Encode hook for JSON serialization of special types.

    Args:
        obj: The object to encode.

    Returns:
        The encoded object as bytes.

    Raises:
        NotImplementedError: If the object type is not supported for encoding.

    Note:
        Currently supports encoding of AttributeDict and HexBytes objects.
    """
    typ = type(obj)
    # sometimes we get a recursion error from the instance checks, this helps us debug that case.
    with reraise_excs_with_extra_context(obj, typ):
        # we use issubclass instead of isinstance here to prevent a recursion error
        if issubclass(typ, int):
            return int(obj)
        elif issubclass(typ, Address):
            return obj[2:]
        elif isinstance(obj, HexBytes32):
            # we trim all leading zeroes since we know how many we need to put back later
            return hex(int(obj.hex(), 16))[2:]
        elif isinstance(obj, HexBytes):
            return bytes(obj).hex()
        elif isinstance(obj, AttributeDict):
            return dict(obj)
        raise TypeError


def dec_hook(typ: Type[T], obj: bytes) -> T:
    """
    Decode hook for JSON deserialization of special types.

    Args:
        typ: The type to decode into.
        obj: The object to decode.

    Returns:
        The decoded object of type T.

    Raises:
        ValueError: If the type is not supported for decoding.

    Note:
        Currently only supports decoding of HexBytes objects.
    """
    if typ is HexBytes:
        return typ(obj)
    raise ValueError(f"{typ} is not a valid type for decoding")


class DiskCache(Generic[S, M], metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _set_metadata(self, from_block: int, done_thru: int) -> None:
        """
        Updates the cache metadata to indicate the cache is populated from block `from_block` to block `to_block`.

        Args:
            from_block: The starting block number.
            done_thru: The ending block number.
        """

    @abc.abstractmethod
    def _is_cached_thru(self, from_block: int) -> int:
        """
        Returns max cached block for this cache or 0 if not cached.

        Args:
            from_block: The starting block number.

        Returns:
            The maximum cached block number.
        """

    @abc.abstractmethod
    def _select(self, from_block: int, to_block: int) -> List[S]:
        """
        Selects all cached objects from block `from_block` to block `to_block`.

        Args:
            from_block: The starting block number.
            to_block: The ending block number.

        Returns:
            A list of cached objects.
        """

    @retry_locked
    def set_metadata(self, from_block: int, done_thru: int) -> None:
        """
        Updates the cache metadata to indicate the cache is populated from block `from_block` to block `to_block`.

        Args:
            from_block: The starting block number.
            done_thru: The ending block number.
        """
        try:
            with db_session:
                return self._set_metadata(from_block, done_thru)
        except TransactionIntegrityError as e:
            logger.debug("%s got exc %s when setting cache metadata", self, e)
            return self.set_metadata(from_block, done_thru)
        except OptimisticCheckError as e:
            # Don't need to update in this case
            logger.debug("%s got exc %s when setting cache metadata", self, e)

    @db_session
    @retry_locked
    def select(self, from_block: int, to_block: int) -> List[S]:
        """
        Selects all cached objects from block `from_block` to block `to_block`.

        Args:
            from_block: The starting block number.
            to_block: The ending block number.

        Returns:
            A list of cached objects.
        """
        return self._select(from_block, to_block)

    @db_session
    @retry_locked
    def is_cached_thru(self, from_block: int) -> int:
        """
        Returns max cached block for this cache or 0 if not cached.

        Args:
            from_block: The starting block number.

        Returns:
            The maximum cached block number.
        """
        return self._is_cached_thru(from_block)

    def check_and_select(self, from_block: int, to_block: int) -> List[S]:
        """
        Selects all cached objects within a specified block range.

        Args:
            from_block: The starting block number.
            to_block: The ending block number.

        Returns:
            A list of cached objects.

        Raises:
            CacheNotPopulatedError: If the cache is not fully populated.
        """
        if self.is_cached_thru(from_block) >= to_block:
            return self.select(from_block, to_block)
        else:
            raise CacheNotPopulatedError(self, from_block, to_block)

    __slots__ = []


C = TypeVar("C", bound=DiskCache)


class _DiskCachedMixin(a_sync.ASyncIterable[T], Generic[T, C], metaclass=abc.ABCMeta):
    __slots__ = "is_reusable", "_cache", "_executor", "_objects", "_pruned"

    def __init__(
        self,
        executor: Optional[_AsyncExecutorMixin] = None,
        is_reusable: bool = True,
    ):
        self.is_reusable = is_reusable
        self._cache = None
        self._executor = executor
        self._objects: List[T] = []
        self._pruned = 0

    @abc.abstractproperty
    def cache(self) -> C: ...

    @property
    def executor(self) -> _AsyncExecutorMixin:
        if self._executor is None:
            self._executor = default_filter_threads
        return self._executor

    @abc.abstractproperty
    def insert_to_db(self) -> Callable[[T], None]: ...

    # abc.abstractproperty
    def bulk_insert(self) -> Callable[[List[T]], Awaitable[None]]: ...

    async def _extend(self, objs: Container[T]) -> None:
        """
        Override this to pre-process objects before storing.

        Args:
            objs: The objects to extend the list with.
        """
        if objs:
            return self._objects.extend(objs)

    def _remove(self, obj: T) -> None:
        self._objects.remove(obj)
        self._pruned += 1

    async def _load_cache(self, from_block: int) -> int:
        """
        Loads cached logs from disk.

        Args:
            from_block: The starting block number.

        Returns:
            The maximum block number loaded from cache.
        """
        logger.debug("checking to see if %s is cached in local db", self)
        if cached_thru := await self.executor.run(
            self.cache.is_cached_thru, from_block
        ):
            logger.info(
                "%s is cached thru block %s, loading from db", self, cached_thru
            )
            await self._extend(
                await self.executor.run(self.cache.select, from_block, cached_thru)
            )
            logger.info(
                "%s loaded %s objects thru block %s from disk",
                self,
                len(self._objects),
                cached_thru,
            )
            return cached_thru
        return None


_E = TypeVar("_E", bound=_AsyncExecutorMixin)
_MAX_LONG_LONG = 9223372036854775807


class Filter(_DiskCachedMixin[T, C]):
    # defaults are stored as class vars to keep instance dicts smaller
    _chunk_size = BATCH_SIZE
    _chunks_per_batch = None
    _exc = None
    _db_task = None
    _sleep_fut = None
    _sleep_time = 60
    _task = None
    _semaphore = None
    _verbose = False
    __slots__ = (
        "from_block",
        "to_block",
        "_interval",
        "_lock",
        "__dict__",
        "__weakref__",
    )

    def __init__(
        self,
        from_block: int,
        *,
        chunk_size: int = BATCH_SIZE,
        chunks_per_batch: Optional[int] = None,
        sleep_time: int = 60,
        semaphore: Optional[dank_mids.BlockSemaphore] = None,
        executor: Optional[_AsyncExecutorMixin] = None,
        is_reusable: bool = True,
        verbose: bool = False,
    ):
        self.from_block = from_block
        if chunk_size != self._chunk_size:
            self._chunk_size = chunk_size
        if chunks_per_batch != self._chunks_per_batch:
            self._chunks_per_batch = chunks_per_batch
        self._lock = a_sync.CounterLock()
        if semaphore != self._semaphore:
            self._semaphore = semaphore
        if sleep_time != self._sleep_time:
            self._sleep_time = sleep_time
        if verbose != self._verbose:
            self._verbose = verbose
        super().__init__(executor=executor, is_reusable=is_reusable)

    def __aiter__(self) -> AsyncIterator[T]:
        return self._objects_thru(block=None).__aiter__()

    def __del__(self) -> None:
        if self._task and not self._task.done():
            self._task.cancel()

    @abc.abstractmethod
    async def _fetch_range(self, from_block: int, to_block: int) -> List[T]: ...

    @property
    def semaphore(self) -> dank_mids.BlockSemaphore:
        if self._semaphore is None:
            self._semaphore = dank_mids.BlockSemaphore(self._chunks_per_batch)
        return self._semaphore

    @property
    def is_asleep(self) -> bool:
        return not self._sleep_fut.done() if self._sleep_fut else False

    def _get_block_for_obj(self, obj: T) -> int:
        """
        Override this as needed for different object types.

        Args:
            obj: The object to get the block number for.

        Returns:
            The block number of the object.
        """
        return obj.blockNumber

    @a_sync.ASyncIterator.wrap
    async def _objects_thru(self, block: Optional[int]) -> AsyncIterator[T]:
        self._ensure_task()
        debug_logs = logger.isEnabledFor(logging.DEBUG)
        yielded = 0
        done_thru = 0
        while True:
            if block is None or done_thru < block:
                if self.is_asleep:
                    self._wakeup()
                await self._lock.wait_for(done_thru + 1)
            if self._exc:
                # create a new duplicate exc instead of building a massive traceback on the original
                try:
                    raise type(self._exc)(*self._exc.args).with_traceback(self._tb)
                except TypeError:
                    raise self._exc.with_traceback(self._tb) from None
            if to_yield := self._objects[yielded - self._pruned :]:
                for obj in to_yield:
                    if block and self._get_block_for_obj(obj) > block:
                        return
                    if not self.is_reusable:
                        self._remove(obj)
                    yield obj
                    yielded += 1
            elif block and done_thru >= block:
                return
            done_thru = self._lock.value
            if debug_logs:
                logger._log(
                    logging.DEBUG,
                    "%s lock value %s to_block %s",
                    (self, done_thru, block),
                )
            if block is None:
                await asyncio.sleep(self._sleep_time)

    @async_property
    async def _sleep(self) -> None:
        if self._sleep_fut is None or self._sleep_fut.done():
            self._sleep_fut = asyncio.get_event_loop().create_future()
        await self._sleep_fut

    def _wakeup(self) -> None:
        self._sleep_fut.set_result(None)

    async def __fetch(self) -> NoReturn:
        try:
            await self._fetch()
        except Exception as e:
            import traceback

            logger.exception(e)
            self._exc = e
            self._tb = e.__traceback__
            # no need to hold vars in memory
            traceback.clear_frames(self._tb)
            self._lock.set(_MAX_LONG_LONG)
            raise

    async def _fetch(self) -> NoReturn:
        """
        Override this if you want.
        """
        await self._loop(self.from_block)

    @stuck_coro_debugger
    async def _fetch_range_wrapped(
        self, i: int, range_start: int, range_end: int, debug_logs: bool
    ) -> List[T]:
        if debug_logs:
            async with self.semaphore[range_end]:
                logger._log(
                    logging.DEBUG,
                    "fetching %s block %s to %s",
                    (self, range_start, range_end),
                )
                return i, range_end, await self._fetch_range(range_start, range_end)
        else:
            async with self.semaphore[range_end]:
                return i, range_end, await self._fetch_range(range_start, range_end)

    async def _loop(self, from_block: int) -> NoReturn:
        logger.debug("starting work loop for %s", self)
        if cached_thru := await self._load_cache(from_block):
            self._lock.set(cached_thru)
        while True:
            await self._load_new_objects(start_from_block=cached_thru or from_block)
            await self._sleep

    @eth_retry.auto_retry
    @stuck_coro_debugger
    async def _load_new_objects(
        self, to_block: Optional[int] = None, start_from_block: Optional[int] = None
    ) -> None:
        SLEEP_TIME = 1

        if debug_logs := logger.isEnabledFor(logging.DEBUG):
            logger._log(logging.DEBUG, "loading new objects for %s", (self,))

        start = (
            v + 1 if (v := self._lock.value) else start_from_block or self.from_block
        )
        if to_block:
            end = to_block
            if start > end:
                raise ValueError(
                    f"start {start} is bigger than end {end}, can't do that"
                )

        elif debug_logs:
            while start > (end := await dank_mids.eth.block_number):
                logger._log(
                    logging.DEBUG,
                    "%s start %s is greater than end %s, sleeping...",
                    (self, start, end),
                )
                await asyncio.sleep(SLEEP_TIME)

        else:
            while start > (end := await dank_mids.eth.block_number):
                await asyncio.sleep(SLEEP_TIME)

        await self._load_range(start, end)

    @stuck_coro_debugger
    async def _load_range(self, from_block: int, to_block: int) -> None:
        if debug_logs := logger.isEnabledFor(logging.DEBUG):
            logger._log(
                logging.DEBUG, "loading block range %s to %s", (from_block, to_block)
            )
        chunks_yielded = 0
        done = {}
        coros = [
            self._fetch_range_wrapped(i, start, end, debug_logs)
            for i, (start, end) in enumerate(
                block_ranges(from_block, to_block, self._chunk_size)
            )
            if self._chunks_per_batch is None or i < self._chunks_per_batch
        ]
        async for i, end, objs in a_sync.as_completed(
            coros, aiter=True, tqdm=self._verbose
        ):
            next_chunk_loaded = False
            done[i] = end, objs
            for i in range(chunks_yielded, len(coros)):
                if i not in done:
                    break
                end, objs = done.pop(i)
                self._insert_chunk(objs, from_block, end, debug_logs)
                await self._extend(objs)
                next_chunk_loaded = True
                chunks_yielded += 1
            if next_chunk_loaded:
                await self._set_lock(end)
                if debug_logs:
                    logger._log(logging.DEBUG, "%s loaded thru block %s", (self, end))

    @stuck_coro_debugger
    async def _set_lock(self, block: int) -> None:
        """
        Override this if you want to, for things like awaiting for tasks to complete as I do in the curve module.

        Args:
            block: The block number to set the lock to.
        """
        self._lock.set(block)

    def _insert_chunk(
        self, objs: List[T], from_block: int, done_thru: int, debug_logs: bool
    ) -> None:
        if (
            (prev_task := self._db_task)
            and prev_task.done()
            and (e := prev_task.exception())
        ):
            raise e
        depth = prev_task._depth + 1 if prev_task else 0
        insert_coro = self.__insert_chunk(
            objs, from_block, done_thru, prev_task, depth, debug_logs
        )
        if debug_logs:
            logger._log(
                logging.DEBUG,
                "%s queuing next db insert chunk %s thru block %s",
                (self, depth, done_thru),
            )
            task = asyncio.create_task(
                coro=insert_coro,
                name=f"_insert_chunk from {from_block} to {done_thru}",
            )
        else:
            task = asyncio.create_task(insert_coro)

        task._depth = depth
        task._prev_task = prev_task
        self._db_task = task

    def _ensure_task(self) -> None:
        if self._task is None:
            logger.debug("creating task for %s", self)
            self._task = asyncio.create_task(
                coro=self.__fetch(), name=f"{self}.__fetch"
            )
            # NOTE: The task does not return and will be cancelled when this object is
            # garbage collected so there is no need to log the "destroy pending task" message.
            self._task._log_destroy_pending = False
        if self._task.done() and (e := self._task.exception()):
            raise e.with_traceback(e.__traceback__)

    @stuck_coro_debugger
    async def __insert_chunk(
        self,
        objs: List[T],
        from_block: int,
        done_thru: int,
        prev_chunk_task: Optional[asyncio.Task],
        depth: int,
        debug_logs: bool,
    ) -> None:
        if prev_chunk_task:
            await prev_chunk_task
        if objs:
            await self.bulk_insert(objs)
        await self.executor.run(self.cache.set_metadata, from_block, done_thru)
        if debug_logs:
            logger._log(
                logging.DEBUG,
                "%s chunk %s thru block %s is now in db",
                (self, depth, done_thru),
            )


def _clean_addresses(addresses) -> Union[str, List[str]]:
    if addresses == ZERO_ADDRESS:
        raise ValueError("Cannot make a LogFilter for the zero address")
    if not addresses:
        return addresses
    if isinstance(addresses, str):
        return convert.to_address(addresses)
    elif hasattr(addresses, "__iter__"):
        if ZERO_ADDRESS in addresses:
            raise ValueError("Cannot make a LogFilter for the zero address")
        return [convert.to_address(address) for address in addresses]
    return convert.to_address(addresses)



import asyncio
import logging
from decimal import Decimal
from typing import List, Optional

import a_sync

from y import ENVIRONMENT_VARIABLES as ENVS
from y.classes.common import ERC20, WeiBalance
from y.contracts import Contract, has_methods
from y.datatypes import AddressOrContract, AnyAddressType, Block, UsdPrice
from y.utils.raw_calls import raw_call

logger = logging.getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
async def is_eps_rewards_pool(token_address: AnyAddressType) -> bool:
    return await has_methods(
        token_address,
        (
            "lpStaker()(address)",
            "rewardTokens(uint)(address)",
            "rewardPerToken(address)(uint)",
            "minter()(address)",
        ),
        sync=False,
    )


@a_sync.a_sync(default="sync")
async def get_price(
    token_address: AddressOrContract,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdPrice:
    minter = await raw_call(
        token_address, "minter()", output="address", block=block, sync=False
    )
    minter = await Contract.coroutine(minter)
    balances: List[WeiBalance]
    i, balances = 0, []
    while True:
        try:
            coin, balance = await asyncio.gather(
                minter.coins.coroutine(i, block_identifier=block),
                minter.balances.coroutine(i, block_identifier=block),
            )
            balance /= await ERC20(coin, asynchronous=True).scale
            balances.append(WeiBalance(balance, coin, block, skip_cache=skip_cache))
            i += 1
        except:
            break
    tvl, total_supply = await asyncio.gather(
        WeiBalance.value_usd.sum(balances, sync=False),
        ERC20(token_address, asynchronous=True).total_supply_readable(block),
    )
    return UsdPrice(tvl / Decimal(total_supply))



import asyncio
import logging
from typing import List, Optional

import a_sync
from brownie import chain

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20
from y.contracts import has_method, has_methods
from y.datatypes import (
    Address,
    AddressOrContract,
    AnyAddressType,
    Block,
    UsdPrice,
    UsdValue,
)
from y.networks import Network
from y.prices import magic
from y.utils.multicall import multicall_same_func_same_contract_different_inputs

logger = logging.getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
async def is_saddle_lp(token_address: AnyAddressType) -> bool:
    pool = await get_pool(token_address, sync=False)
    return pool is not None and await has_methods(
        pool,
        ("getVirtualPrice()(uint)", "getA()(uint)", "getAPrecise()(uint)"),
        sync=False,
    )


@a_sync.a_sync(default="sync", ram_cache_ttl=ENVS.CACHE_TTL)
async def get_pool(token_address: AnyAddressType) -> Address:
    token_address = await convert.to_address_async(token_address)
    if chain.id == Network.Mainnet:
        if (
            token_address == "0xc9da65931ABf0Ed1b74Ce5ad8c041C4220940368"
        ):  # saddle aleth doesn't have swap() function
            return "0xa6018520EAACC06C30fF2e1B3ee2c7c22e64196a"
        elif token_address == "0xd48cF4D7FB0824CC8bAe055dF3092584d0a1726A":  # saddle d4
            return "0xC69DDcd4DFeF25D8a793241834d4cc4b3668EAD6"
        elif token_address == "0xF32E91464ca18fc156aB97a697D6f8ae66Cd21a3":
            return "0xdf3309771d2BF82cb2B6C56F9f5365C8bD97c4f2"
    pool = await has_method(
        token_address, "swap()(address)", return_response=True, sync=False
    )
    return pool or None


@a_sync.a_sync(default="sync")
async def get_price(
    token_address: AddressOrContract, block: Optional[Block] = None
) -> UsdPrice:
    tvl, total_supply = await asyncio.gather(
        get_tvl(token_address, block, sync=False),
        ERC20(token_address, asynchronous=True).total_supply_readable(block),
    )
    return UsdPrice(tvl / total_supply)


@a_sync.a_sync(default="sync")
async def get_tvl(
    token_address: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> UsdValue:
    tokens: List[ERC20]
    pool, tokens, balances = await asyncio.gather(
        get_pool(token_address, sync=False),
        get_tokens(token_address, block, sync=False),
        multicall_same_func_same_contract_different_inputs(
            pool,
            "getTokenBalance(uint8)(uint)",
            inputs=[*range(len(tokens))],
            sync=False,
        ),
    )
    scales, prices = await a_sync.gather(
        ERC20.scale.map(tokens).values(pop=True),
        magic.get_prices(tokens, block, skip_cache=skip_cache, silent=True, sync=False),
    )
    return UsdValue(
        sum(
            balance / scale * price
            for balance, scale, price in zip(balances, scales, prices)
        )
    )


@a_sync.a_sync(default="sync")
async def get_tokens(
    token_address: AnyAddressType, block: Optional[Block] = None
) -> List[ERC20]:
    pool, response = await asyncio.gather(
        get_pool(token_address, sync=False),
        multicall_same_func_same_contract_different_inputs(
            pool,
            "getToken(uint8)(address)",
            inputs=[*range(8)],
            block=block,
            return_None_on_failure=True,
            sync=False,
        ),
    )
    return [ERC20(token) for token in response if token is not None]



import asyncio
import logging
import threading
from contextlib import suppress
from typing import List, Optional, Tuple, Union

import a_sync
from brownie import ZERO_ADDRESS, chain
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y._decorators import stuck_coro_debugger
from y.classes.common import ERC20
from y.datatypes import Address, AnyAddressType, Block, Pool, UsdPrice
from y.exceptions import NonStandardERC20, contract_not_verified
from y.networks import Network
from y.prices.dex.solidly import SolidlyRouter
from y.prices.dex.uniswap.v1 import UniswapV1
from y.prices.dex.uniswap.v2 import NotAUniswapV2Pool, UniswapRouterV2, UniswapV2Pool
from y.prices.dex.uniswap.v2_forks import UNISWAPS
from y.prices.dex.uniswap.v3 import UniswapV3, uniswap_v3
from y.prices.dex.velodrome import VelodromeRouterV2
from y.utils.logging import _gh_issue_request, get_price_logger

logger = logging.getLogger(__name__)

# NOTE: If this is failing to pull a price for a token you need, it's likely because that token requires a special swap path.
#       Please add a viable swap path to ..protocols to fetch price data successfully.

_special_routers = {
    "solidly": SolidlyRouter,
    "velodrome v1": SolidlyRouter,
    "velodrome v2": VelodromeRouterV2,
    "aerodrome": VelodromeRouterV2,
    "ramses": SolidlyRouter,
}

Uniswap = Union[UniswapV1, UniswapRouterV2, UniswapV3]


class UniswapMultiplexer(a_sync.ASyncGenericSingleton):
    """
    A multiplexer for Uniswap routers that provides aggregated functionality across multiple Uniswap instances and/or forks.
    """

    def __init__(self, *, asynchronous: bool = False) -> None:
        super().__init__()
        self.asynchronous = asynchronous
        self.v2_routers = {}
        for name in UNISWAPS:
            router_cls = _special_routers.get(name, UniswapRouterV2)
            try:
                self.v2_routers[name] = router_cls(
                    UNISWAPS[name]["router"], asynchronous=self.asynchronous
                )
            except ValueError as e:  # TODO do this better
                if not contract_not_verified(e):
                    raise
        self.v1 = (
            UniswapV1(asynchronous=self.asynchronous)
            if chain.id == Network.Mainnet
            else None
        )
        self.v3 = UniswapV3(asynchronous=self.asynchronous) if uniswap_v3 else None

        self.uniswaps: List[Uniswap] = []
        if self.v1:
            self.uniswaps.append(self.v1)
        self.uniswaps.extend(self.v2_routers.values())
        if self.v3:
            self.uniswaps.append(self.v3)

        self.v2_factories = [UNISWAPS[name]["factory"] for name in UNISWAPS]
        self._uid_lock = threading.Lock()

    @stuck_coro_debugger
    async def is_uniswap_pool(self, token_address: AnyAddressType) -> bool:
        token_address = await convert.to_address_async(token_address)
        try:
            await ERC20(token_address, asynchronous=True).decimals
        except NonStandardERC20:
            return False

        pool = UniswapV2Pool(token_address, asynchronous=True)
        with suppress(NotAUniswapV2Pool, ContractLogicError):
            if await pool.is_uniswap_pool(sync=False):
                factory = await pool.__factory__
                if factory not in self.v2_factories and factory != ZERO_ADDRESS:
                    _gh_issue_request(
                        f"UniClone Factory {factory} is unknown to ypricemagic.", logger
                    )
                    self.v2_factories.append(factory)
                return True
        return False

    @stuck_coro_debugger
    async def get_price(
        self,
        token_in: AnyAddressType,
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
        skip_cache: bool = ENVS.SKIP_CACHE,
    ) -> Optional[UsdPrice]:
        """
        Calculate a price based on Uniswap Router quote for selling one `token_in`.
        Always finds the deepest swap path for `token_in`.
        """
        router: Uniswap
        token_in = await convert.to_address_async(token_in)
        logger = get_price_logger(token_in, block, extra=type(self).__name__)
        routers_by_depth = await self.routers_by_depth(
            token_in, block=block, ignore_pools=ignore_pools, sync=False
        )
        logger.debug("uniswap routers by depth: %s", routers_by_depth)
        for router in routers_by_depth:
            # tries each known router from most to least liquid
            # returns the first price we get back, almost always from the deepest router
            logger.debug("fetching from %s", router)
            price = await router.get_price(
                token_in,
                block=block,
                ignore_pools=ignore_pools,
                skip_cache=skip_cache,
                sync=False,
            )
            logger.debug("%s -> %s", router, price)
            if price:
                return price

    @stuck_coro_debugger
    async def routers_by_depth(
        self,
        token_in: AnyAddressType,
        block: Optional[Block] = None,
        ignore_pools: Tuple[Pool, ...] = (),
    ) -> List[UniswapRouterV2]:
        """
        Get Uniswap routers sorted by liquidity depth for a given token.

        This method checks the liquidity of the input token across all Uniswap instances
        and returns a list of routers sorted by their liquidity depth in descending order.

        Args:
            token_in: The address of the input token to check liquidity for.
            block (optional): The block number to query. Defaults to latest block.
            ignore_pools (optional): A tuple of Pool objects to ignore when checking liquidity.

        Returns:
            A list of UniswapRouterV2 objects sorted by liquidity depth in descending order.

        Note:
            - The method uses asyncio.gather to check liquidity across all Uniswap instances concurrently.
            - Routers with zero liquidity are excluded from the result.
        """
        token_in = await convert.to_address_async(token_in)
        depth_to_router = dict(
            zip(
                await asyncio.gather(
                    *[
                        uniswap.check_liquidity(
                            token_in, block, ignore_pools=ignore_pools, sync=False
                        )
                        for uniswap in self.uniswaps
                    ]
                ),
                self.uniswaps,
            )
        )
        return [
            depth_to_router[balance]
            for balance in sorted(depth_to_router, reverse=True)
            if balance
        ]

    @a_sync.Semaphore(100)  # some arbitrary number to keep the loop unclogged
    @stuck_coro_debugger
    async def check_liquidity(
        self,
        token: Address,
        block: Block,
        ignore_pools: Tuple[Pool, ...] = (),
    ) -> int:
        """
        Check the maximum liquidity for a token across all Uniswap instances.

        This method checks the liquidity of the given token across all Uniswap instances
        and returns the maximum liquidity found.

        Args:
            token: The address of the token to check liquidity for.
            block: The block number to query.
            ignore_pools (optional): A tuple of Pool objects to ignore when checking liquidity.

        Returns:
            The maximum liquidity found for the token across all Uniswap instances.

        Note:
            - The method uses asyncio.gather to check liquidity across all Uniswap instances concurrently.
            - A semaphore is used to limit the number of concurrent checks to 100.
        """
        return max(
            await asyncio.gather(
                *[
                    uniswap.check_liquidity(
                        token, block, ignore_pools=ignore_pools, sync=False
                    )
                    for uniswap in self.uniswaps
                ]
            )
        )


uniswap_multiplexer = UniswapMultiplexer(asynchronous=True)



import asyncio
import json
import logging
import threading
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    NewType,
    Optional,
    Set,
    Tuple,
    Union,
)

import a_sync
import dank_mids
import eth_retry
from brownie import ZERO_ADDRESS, chain, web3
from brownie.exceptions import CompilerError, ContractNotFound
from brownie.network.contract import (
    ContractEvents,
    _add_deployment,
    _ContractBase,
    _DeployedContractBase,
    _fetch_from_explorer,
    _resolve_address,
)
from brownie.typing import AccountsType
from checksum_dict import ChecksumAddressDict, ChecksumAddressSingletonMeta
from hexbytes import HexBytes
from multicall import Call
from typing_extensions import Self
from web3.exceptions import ContractLogicError

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert, exceptions
from y._decorators import stuck_coro_debugger
from y.datatypes import Address, AnyAddressType, Block
from y.interfaces.ERC20 import ERC20ABI
from y.networks import Network
from y.time import check_node, check_node_async
from y.utils.cache import memory
from y.utils.events import Events
from y.utils.gather import gather_methods

logger = logging.getLogger(__name__)

_brownie_deployments_db_lock = threading.Lock()

# These tokens have trouble when resolving the implementation via the chain.
FORCE_IMPLEMENTATION = {
    Network.Mainnet: {
        "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48": "0xa2327a938Febf5FEC13baCFb16Ae10EcBc4cbDCF",  # USDC as of 2022-08-10
        "0x3d1E5Cf16077F349e999d6b21A4f646e83Cd90c5": "0xf51fC5ae556F5B8c6dCf50f70167B81ceb02a2b2",  # dETH as of 2024-02-15
    },
}.get(chain.id, {})


def Contract_erc20(address: AnyAddressType) -> "Contract":
    """
    Create a Contract instance for an ERC20 token.

    This function uses the standard ERC20 ABI instead of fetching the contract ABI from the block explorer.

    Args:
        address: The address of the ERC20 token.

    Returns:
        A Contract instance for the ERC20 token.
    """
    address = convert.to_address(address)
    return Contract.from_abi("ERC20", address, ERC20ABI)


def Contract_with_erc20_fallback(address: AnyAddressType) -> "Contract":
    """
    Create a Contract instance for an address, falling back to an ERC20 token if the contract is not verified.

    Args:
        address: The address of the contract or ERC20 token.

    Returns:
        A Contract instance for the contract address.
    """
    if isinstance(address, Contract):
        return address
    address = convert.to_address(address)
    try:
        return Contract(address)
    except exceptions.ContractNotVerified:
        return Contract_erc20(address)


@memory.cache()
# yLazyLogger(logger)
@eth_retry.auto_retry
def contract_creation_block(
    address: AnyAddressType, when_no_history_return_0: bool = False
) -> int:
    """
    Determine the block when a contract was created using binary search.
    NOTE Requires access to historical state. Doesn't account for CREATE2 or SELFDESTRUCT.

    Args:
        address: The address of the contract.
        when_no_history_return_0: If True, return 0 when no history is found instead of raising a :class:`~exceptions.NodeNotSynced` exception. Default False.

    Returns:
        The block number at which the contract was created.

    Raises:
        exceptions.NodeNotSynced: If the node is not fully synced.
        ValueError: If the contract creation block cannot be determined.
    """
    address = convert.to_address(address)
    logger.debug("contract creation block %s", address)
    height = chain.height

    if height == 0:
        raise exceptions.NodeNotSynced(_NOT_SYNCED)

    check_node()

    lo, hi = 0, height
    barrier = 0
    warned = False
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        # TODO rewrite this so we can get deploy blocks for some contracts deployed on correct side of barrier
        try:
            if _get_code(address, mid):
                hi = mid
            else:
                lo = mid
        except ValueError as e:
            if "missing trie node" in str(e) and not warned:
                logger.warning(
                    "missing trie node, `contract_creation_block` may output a higher block than actual. Please try again using an archive node."
                )
            elif "Server error: account aurora does not exist while viewing" in str(e):
                if not warned:
                    logger.warning(str(e))
            elif "No state available for block" in str(e):
                if not warned:
                    logger.warning(str(e))
            else:
                raise
            warned = True
            barrier = mid
            lo = mid
    if hi == lo + 1 == barrier + 1 and when_no_history_return_0:
        logger.warning(
            "could not determine creation block for %s on %s (deployed prior to barrier)",
            address,
            Network.name(),
        )
        logger.debug("contract creation block %s -> 0", address)
        return 0
    if hi != height:
        logger.debug("contract creation block %s -> %s", address, hi)
        return hi
    raise ValueError(f"Unable to find deploy block for {address} on {Network.name()}")


get_code = eth_retry.auto_retry(dank_mids.eth.get_code)


@memory.cache
@eth_retry.auto_retry
def _get_code(address: str, block: int) -> HexBytes:
    return web3.eth.get_code(address, block)


creation_block_semaphore = a_sync.ThreadsafeSemaphore(10)


@a_sync.a_sync(cache_type="memory")
@stuck_coro_debugger
@eth_retry.auto_retry
async def contract_creation_block_async(
    address: AnyAddressType, when_no_history_return_0: bool = False
) -> int:
    """
    Determine the block when a contract was created using binary search.
    NOTE Requires access to historical state. Doesn't account for CREATE2 or SELFDESTRUCT.

    Args:
        address: The address of the contract.
        when_no_history_return_0: If True, return 0 when no history is found instead of raising a :class:`~exceptions.NodeNotSynced` exception. Default False.

    Returns:
        The block number at which the contract was created.

    Raises:
        exceptions.NodeNotSynced: If the node is not fully synced.
        ValueError: If the contract creation block cannot be determined.
    """
    from y._db.utils import contract as db

    address = await convert.to_address_async(address)
    if deploy_block := await db.get_deploy_block(address):
        return deploy_block

    logger.debug(f"contract creation block {address}")
    height = await dank_mids.eth.block_number

    if height == 0:
        raise exceptions.NodeNotSynced(_NOT_SYNCED)

    await check_node_async()

    lo, hi = 0, height
    barrier = 0
    warned = False
    while hi - lo > 1:
        mid = lo + (hi - lo) // 2
        # TODO rewrite this so we can get deploy blocks for some contracts deployed on correct side of barrier
        try:
            if await get_code(address, mid):
                hi = mid
            else:
                lo = mid
        except ValueError as e:
            if "missing trie node" in str(e):
                if not warned:
                    logger.warning(
                        "missing trie node, `contract_creation_block` may output a higher block than actual. Please try again using an archive node."
                    )
            elif "Server error: account aurora does not exist while viewing" in str(e):
                if not warned:
                    logger.warning(str(e))
            elif "No state available for block" in str(e):
                if not warned:
                    logger.warning(str(e))
            else:
                raise
            warned = True
            barrier = mid
            lo = mid
    if hi == lo + 1 == barrier + 1 and when_no_history_return_0:
        logger.warning(
            f"could not determine creation block for {address} on {Network.name()} (deployed prior to barrier)"
        )
        logger.debug(f"contract creation block {address} -> 0")
        return 0
    if hi != height:
        logger.debug(f"contract creation block {address} -> {hi}")
        db.set_deploy_block(address, hi)
        return hi
    raise ValueError(f"Unable to find deploy block for {address} on {Network.name()}")


# this defaultdict prevents congestion in the contracts thread pool
address_semaphores = defaultdict(lambda: a_sync.ThreadsafeSemaphore(1))


class ContractEvents(ContractEvents):
    def __getattr__(self, name: str) -> Events:
        return super().__getattr__(name)


class CompilerError(Exception):
    def __init__(self):
        super().__init__(
            "y.Contract objects work best when we bypass compilers.\n"
            "In this case, it will *only* work when we bypass.\n"
            "Please ensure autofetch_sources=False in your brownie-config.yaml and rerun your script."
        )


class Contract(dank_mids.Contract, metaclass=ChecksumAddressSingletonMeta):
    """
    Though it may look complicated, a ypricemagic Contract object is simply a brownie Contract object with a few modifications:

        1. Contracts will not be compiled. This allows you to more quickly fetch contracts from the block explorer and prevents you from having to download and install compilers.
            NOTE: You must set `autofetch_sources=false` in your project's brownie-config.yaml for this to work correctly.

        2. To each contract method, a `coroutine` property has been defined which allows you to make asynchronous calls which are intelligently batched in the background by :mod:`dank_mids` to reduce overhead.
            Example:
                >>> contract = Contract("0xAddress")
                >>> contract.methodName(*args, block_identifier=12345678)
                1000000000000000000
                >>> coro = contract.methodName.coroutine(*args, block_identifier=12345678)
                >>> coro
                <coroutine coroutine object at 0x12345678>
                >>> contract.methodName(*args, block_identifier=12345678) == await coro
                True

        3. New methods:
            - :meth:`~has_method`
            - :meth:`~has_methods`
            - :meth:`~build_name`
            - :meth:`~get_code`

        4. A few attributes were removed in order to minimize the size of a Contract object in memory:
            - :attr:`~ast`
            - :attr:`~bytecode`
            - :attr:`~coverageMap`
            - :attr:`~deployedBytecode`
            - :attr:`~deployedSourceMap`
            - :attr:`~natspec`
            - :attr:`~opcodes`
            - :attr:`~pcMap`
    """

    # the default state for Contract objects
    verified = True
    """True if the contract is verified on this network's block explorer. False otherwise."""

    events: ContractEvents
    """
    A container for various event types associated with this contract.
    
    Provides a convenient way to query contract events with minimal code.
    """

    _ChecksumAddressSingletonMeta__instances: ChecksumAddressDict["Contract"]
    _ttl_cache_popper: Union[Literal["disabled"], int, asyncio.TimerHandle]

    @eth_retry.auto_retry
    def __init__(
        self,
        address: AnyAddressType,
        owner: Optional[AccountsType] = None,
        require_success: bool = True,
        cache_ttl: Optional[int] = ENVS.CONTRACT_CACHE_TTL,  # units: seconds
    ) -> None:
        """
        Note:
            autofetch-sources: false
        """

        address = str(address)
        if address.lower() in [
            "0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee",
            ZERO_ADDRESS,
        ]:
            raise ContractNotFound(f"{address} is not a contract.")
        if require_success and address in _unverified:
            raise exceptions.ContractNotVerified(address)

        try:
            # Try to fetch the contract from the local sqlite db.
            with _brownie_deployments_db_lock:
                super().__init__(address, owner=owner)
            if not isinstance(self.verified, bool) and self.verified is not None:
                logger.warning(
                    f'`Contract("{address}").verified` property will not be usable due to the contract having a `verified` method in its ABI.'
                )
            self.__finish_init(cache_ttl)
            return
        except (AssertionError, IndexError) as e:
            if str(e) == "pop from an empty deque" or isinstance(e, AssertionError):
                raise CompilerError from None
            raise
        except ValueError as e:
            logger.debug(f"{e}")
            if not str(e).startswith("Unknown contract address: "):
                raise

        # The contract does not exist in your local brownie deployments.db
        try:
            name, abi = _resolve_proxy(address)
            build = {
                "abi": abi,
                "address": address,
                "contractName": name,
                "type": "contract",
            }
            self.__init_from_abi__(build, owner=owner, persist=True)
            self.__finish_init(cache_ttl)
        except (ContractNotFound, exceptions.ContractNotVerified) as e:
            if isinstance(e, exceptions.ContractNotVerified):
                _unverified.add(address)
            if require_success:
                raise
            try:
                if isinstance(e, exceptions.ContractNotVerified):
                    self.verified = False
                    self._build = {"contractName": "Non-Verified Contract"}
                else:
                    self.verified = None
                    self._build = {"contractName": "Broken Contract"}
            except AttributeError:
                logger.warning(
                    f'`Contract("{address}").verified` property will not be usable due to the contract having a `verified` method in its ABI.'
                )

    @classmethod
    @a_sync.a_sync
    def from_abi(
        cls,
        name: str,
        address: str,
        abi: List,
        owner: Optional[AccountsType] = None,
        persist: bool = True,
        cache_ttl: Optional[int] = ENVS.CONTRACT_CACHE_TTL,  # units: seconds
    ) -> Self:
        """
        Create a Contract instance from an ABI.

        Args:
            name: The name of the contract.
            address: The address of the contract.
            abi: The ABI of the contract.
            owner (optional): The owner of the contract. Default None.
            persist (optional): If True, persist the contract in brownie's local contract database. Default True.
            cache_ttl (optional): The time-to-live for the contract cache in seconds. Default set in :mod:`~y.ENVIRONMENT_VARIABLES`.

        Returns:
            A Contract instance for the given ABI.
        """
        self = cls.__new__(cls)
        build = {
            "abi": abi,
            "address": _resolve_address(address),
            "contractName": name,
            "type": "contract",
        }
        self.__init_from_abi__(build, owner, persist)
        self.__finish_init(cache_ttl)
        return self

    @classmethod
    async def coroutine(
        cls,
        address: AnyAddressType,
        require_success: bool = True,
        cache_ttl: Optional[int] = ENVS.CONTRACT_CACHE_TTL,  # units: seconds
    ) -> Self:

        address = str(address)
        if contract := cls._ChecksumAddressSingletonMeta__instances.get(address, None):
            return contract
        # dict lookups faster than string comparisons, keep this behind the singleton check
        if address.lower() in [
            "0xeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeeee",
            ZERO_ADDRESS,
        ]:
            raise ContractNotFound(f"{address} is not a contract.") from None
        try:
            # We do this so we don't clog the threadpool with multiple jobs for the same contract.
            return await _contract_queue(
                address, require_success=require_success, cache_ttl=cache_ttl
            )
        except (ContractNotFound, exceptions._ExplorerError, CompilerError) as e:
            # re-raise with nicer traceback
            raise type(e)(*e.args) from None

    @classmethod
    @stuck_coro_debugger
    async def _coroutine(
        cls,
        address: AnyAddressType,
        require_success: bool = True,
        cache_ttl: Optional[int] = ENVS.CONTRACT_CACHE_TTL,  # units: seconds
    ) -> Self:
        """
        Internal method to create a Contract instance asynchronously.

        Args:
            address: The address of the contract.
            require_success: If True, raise an exception if the contract cannot be initialized.
            cache_ttl: The time-to-live for the contract cache in seconds.

        Returns:
            A Contract instance for the given address.
        """
        contract = await ENVS.CONTRACT_THREADS.run(
            cls, address, require_success=require_success
        )

        if not contract.verified or contract._ttl_cache_popper == "disabled":
            pass

        elif cache_ttl is None:
            if isinstance(contract._ttl_cache_popper, asyncio.TimerHandle):
                contract._ttl_cache_popper.cancel()
            contract._ttl_cache_popper = "disabled"

        elif isinstance(contract._ttl_cache_popper, int):
            cache_ttl = max(contract._ttl_cache_popper, cache_ttl)
            contract._ttl_cache_popper = asyncio.get_running_loop().call_later(
                cache_ttl,
                cls._ChecksumAddressSingletonMeta__instances.pop,
                contract.address,
                None,
            )

        elif (
            asyncio.get_running_loop().time() + cache_ttl
            > contract._ttl_cache_popper.when()
        ):
            contract._ttl_cache_popper.cancel()
            contract._ttl_cache_popper = asyncio.get_running_loop().call_later(
                cache_ttl,
                cls._ChecksumAddressSingletonMeta__instances.pop,
                contract.address,
                None,
            )
        return contract

    def __init_from_abi__(
        self, build: Dict, owner: Optional[AccountsType] = None, persist: bool = True
    ) -> None:
        """
        Initialize a Contract instance from an ABI.

        Args:
            build: The build information for the contract.
            owner (optional): The owner of the contract. Default None.
            persist (optional): If True, persist the contract in the local database. Default True.

        Returns:
            The initialized Contract instance.
        """
        _ContractBase.__init__(self, None, build, {})  # type: ignore
        _DeployedContractBase.__init__(self, build["address"], owner, None)
        if persist:
            _add_deployment(self)
        try:
            self.verified = True
        except AttributeError:
            logger.warning(
                f'`Contract("{self.address}").verified` property will not be usable due to the contract having a `verified` method in its ABI.'
            )
        return self

    def has_method(
        self, method: str, return_response: bool = False
    ) -> Union[bool, Any]:
        """
        Check if the contract has a specific method.

        Args:
            method: The name of the method to check for.
            return_response (optional): If True, return the response of the method call instead of a boolean. Default False.

        Returns:
            A boolean indicating whether the contract has the method, or the response of the method call if return_response is True.
        """
        return has_method(
            self.address, method, return_response=return_response, sync=False
        )

    async def has_methods(
        self, methods: List[str], _func: Union[any, all] = all
    ) -> bool:
        """
        Check if the contract has all the specified methods.

        Args:
            methods: A list of method names to check for.
            _func (optional): The function to use for combining the results (either :func:`all` or :func:`any`). Default :func:`all`.

        Returns:
            A boolean indicating whether the contract has all the specified methods.
        """
        return await has_methods(self.address, methods, _func, sync=False)

    async def build_name(self, return_None_on_failure: bool = False) -> Optional[str]:
        """
        Get the build name of the contract.

        Args:
            return_None_on_failure (optional): If True, return None if the build name cannot be determined instead of raising an Exception. Default False.

        Returns:
            The build name of the contract, or None if the build name cannot be determined and return_None_on_failure is True.
        """
        return await build_name(
            self.address, return_None_on_failure=return_None_on_failure, sync=False
        )

    async def get_code(self, block: Optional[Block] = None) -> HexBytes:
        """
        Get the bytecode of the contract at a specific block.

        Args:
            block (optional): The block number at which to get the bytecode. Defaults to latest block.

        Returns:
            The bytecode of the contract at the specified block.
        """
        return await get_code(self.address, block=block)

    def _schedule_cache_pop(self, cache_ttl: Optional[int]) -> None:
        if cache_ttl is None:
            self._ttl_cache_popper = "disabled"
            return

        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # If the event loop isn't running yet we can just specify the cache_ttl for later use
            self._ttl_cache_popper = cache_ttl
            return

        cache = self._ChecksumAddressSingletonMeta__instances

        self._ttl_cache_popper = loop.call_later(
            cache_ttl,
            cache.pop,
            self.address,
            None,
        )

    def __finish_init(self, cache_ttl):
        # Patch the Contract with coroutines for each method.
        # TODO I think we can maybe remove this now, gotta check
        dank_mids.patch_contract(self)

        # Init an event container for each topic
        _setup_events(self)

        # Get rid of unnecessary memory-hog properties
        _squeeze(self)

        # schedule call to pop from cache
        self._schedule_cache_pop(cache_ttl)


_contract_queue = a_sync.SmartProcessingQueue(Contract._coroutine, num_workers=32)


@memory.cache()
# TODO: async this and put it into ydb for quicker startups
# yLazyLogger(logger)
def is_contract(address: AnyAddressType) -> bool:
    """
    Checks to see if the input address is a contract. Returns `False` if:
    - The address is not and has never been a contract
    - The address used to be a contract but has self-destructed

    Args:
        address: The address to check.

    Returns:
        True if the address is a contract, False otherwise.
    """
    address = convert.to_address(address)
    return web3.eth.get_code(address) not in ["0x", b""]


@a_sync.a_sync(default="sync", cache_type="memory")
async def has_method(
    address: Address, method: str, return_response: bool = False
) -> Union[bool, Any]:
    """
    Checks to see if a contract has a `method` view method with no inputs.
    `return_response=True` will return `response` in bytes if `response` else `False`

    Args:
        address: The address of the contract.
        method: The name of the method to check for.
        return_response: If True, return the response of the method call instead of a boolean. Default False.

    Returns:
        A boolean indicating whether the contract has the method, or the response of the method call if return_response is True.
    """
    address = await convert.to_address_async(address)
    try:
        response = await Call(address, [method])
        return False if response is None else response if return_response else True
    except Exception as e:
        if (
            isinstance(e, ContractLogicError)
            or exceptions.call_reverted(e)
            or "invalid jump destination" in str(e)
        ):
            return False
        raise


@stuck_coro_debugger
@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=15 * 60)
async def has_methods(
    address: AnyAddressType,
    methods: Tuple[str],
    _func: Callable = all,  # Union[any, all]
) -> bool:
    """
    Checks to see if a contract has each view method (with no inputs) in `methods`.
    Pass `at_least_one=True` to only verify a contract has at least one of the methods.

    Args:
        address: The address of the contract.
        methods: A tuple of method names to check for.
        _func: The function to use for combining the results (either :func:`all` or :func:`any`).

    Returns:
        A boolean indicating whether the contract has all the specified methods.
    """
    assert _func in [all, any], "`_func` must be either `any` or `all`"

    address = await convert.to_address_async(address)
    try:
        return _func(
            [result is not None for result in await gather_methods(address, methods)]
        )
    except Exception as e:
        if not isinstance(e, ContractLogicError) and not exceptions.call_reverted(e):
            raise  # and not out_of_gas(e): raise
        # Out of gas error implies one or more method is state-changing.
        # If `_func == all` we return False because `has_methods` is only supposed to work for public view methods with no inputs
        # If `_func == any` maybe one of the methods will work without "out of gas" error
        return (
            False
            if _func == all
            else any(
                await asyncio.gather(
                    *[has_method(address, method, sync=False) for method in methods]
                )
            )
        )


# yLazyLogger(logger)
@stuck_coro_debugger
async def probe(
    address: AnyAddressType,
    methods: List[str],
    block: Optional[Block] = None,
    return_method: bool = False,
) -> Any:
    address = await convert.to_address_async(address)
    results = await gather_methods(
        address, methods, block=block, return_exceptions=True
    )
    logger.debug("probe results: %s", results)
    results = [
        (method, result)
        for method, result in zip(methods, results)
        if not isinstance(result, Exception) and result is not None
    ]
    if len(results) not in [1, 0]:
        logger.debug("multiple results: %s", results)
        if len(results) != 2 or results[0][1] != results[1][1]:
            raise AssertionError(
                f"`probe` returned multiple results for {address}: {results}. Must debug"
            )
        method = results[0][0], results[1][0]
        result = results[0][1]
        results = [(method, result)]
        logger.debug("final results: %s", results)
    method, result = results[0] if len(results) == 1 else (None, None)
    if method:
        assert result is not None
    return (method, result) if return_method else result


@a_sync.a_sync(default="sync")
@stuck_coro_debugger
async def build_name(
    address: AnyAddressType, return_None_on_failure: bool = False
) -> str:
    """
    Get the build name of a contract.

    Args:
        address: The address of the contract.
        return_None_on_failure (optional): If True, return None if the build name cannot be determined instead of raising an Exception. Default False.

    Returns:
        The build name of the contract, or None if the build name cannot be determined and return_None_on_failure is True.
    """
    try:
        contract = await Contract.coroutine(address)
        return contract.__dict__["_build"]["contractName"]
    except exceptions.ContractNotVerified:
        if not return_None_on_failure:
            raise
        return None


async def proxy_implementation(
    address: AnyAddressType, block: Optional[Block]
) -> Address:
    """
    Get the implementation address for a proxy contract.

    Args:
        address: The address of the proxy contract.
        block: The block number at which to get the implementation address. Defaults to latest block.

    Returns:
        The address of the implementation contract.
    """
    return await probe(
        address, ["implementation()(address)", "target()(address)"], block
    )


def _squeeze(contract: Contract) -> Contract:
    """
    Reduce the contract size in RAM by removing large data structures from the build.

    Args:
        contract: The contract object to squeeze.

    Returns:
        The squeezed contract object.
    """
    for k in [
        "ast",
        "bytecode",
        "coverageMap",
        "deployedBytecode",
        "deployedSourceMap",
        "natspec",
        "opcodes",
        "pcMap",
    ]:
        if contract._build and k in contract._build.keys():
            contract._build[k] = {}
    return contract


@eth_retry.auto_retry
def _extract_abi_data(address: Address):
    """
    Extract ABI data for a contract from the blockchain explorer.

    Args:
        address: The address of the contract.

    Returns:
        A tuple containing the contract name, ABI, and implementation address (if applicable).

    Raises:
        Various exceptions based on the API response and contract status.
    """
    try:
        data = _fetch_from_explorer(address, "getsourcecode", False)["result"][0]
    except ConnectionError as e:
        if '{"message":"Something went wrong.","result":null,"status":"0"}' in str(e):
            if chain.id == Network.xDai:
                raise ValueError("Rate limited by Blockscout. Please try again.") from e
            if web3.eth.get_code(address):
                raise exceptions.ContractNotVerified(address) from e
            else:
                raise ContractNotFound(address) from e
        raise
    except ValueError as e:
        if (
            str(e).startswith("Failed to retrieve data from API")
            and "invalid api key" in str(e).lower()
        ):
            raise exceptions.InvalidAPIKeyError from e
        if exceptions.contract_not_verified(e):
            raise exceptions.ContractNotVerified(
                f"{address} on {Network.printable()}"
            ) from e
        elif "Unknown contract address:" in str(e):
            exc_type = (
                exceptions.ContractNotVerified
                if is_contract(address)
                else ContractNotFound
            )
            raise exc_type(str(e)) from e
        else:
            raise

    is_verified = bool(data.get("SourceCode"))
    if not is_verified:
        raise exceptions.ContractNotVerified(
            f"Contract source code not verified: {address}"
        ) from None
    name = data["ContractName"]
    abi = json.loads(data["ABI"])
    implementation = data.get("Implementation")
    return name, abi, implementation


def _resolve_proxy(address) -> Tuple[str, List]:
    """
    Resolve the implementation address for a proxy contract.

    Args:
        address: The address of the proxy contract.

    Returns:
        A tuple containing the contract name and ABI.
    """
    address = convert.to_address(address)
    name, abi, implementation = _extract_abi_data(address)
    as_proxy_for = None

    if address in FORCE_IMPLEMENTATION:
        implementation = FORCE_IMPLEMENTATION[address]
        name, implementation_abi, _ = _extract_abi_data(implementation)
        # Here we merge the proxy ABI with the implementation ABI
        # without doing this, we'd only get the implementation
        # and would lack any valid methods/events from the proxy itself.
        # Credit: Wavey@Yearn
        abi += implementation_abi
        return name, abi

    # always check for an EIP1967 proxy - https://eips.ethereum.org/EIPS/eip-1967
    implementation_eip1967 = web3.eth.get_storage_at(
        address, int(web3.keccak(text="eip1967.proxy.implementation").hex(), 16) - 1
    )
    # always check for an EIP1822 proxy - https://eips.ethereum.org/EIPS/eip-1822
    implementation_eip1822 = web3.eth.get_storage_at(
        address, web3.keccak(text="PROXIABLE")
    )

    # Just leave this code where it is for a helpful debugger as needed.
    if address == "":
        raise Exception(
            f"""implementation: {implementation}
            implementation_eip1967: {len(implementation_eip1967)} {implementation_eip1967}
            implementation_eip1822: {len(implementation_eip1822)} {implementation_eip1822}"""
        )

    if len(implementation_eip1967) > 0 and int(implementation_eip1967.hex(), 16):
        as_proxy_for = _resolve_address(implementation_eip1967[-20:])
    elif len(implementation_eip1822) > 0 and int(implementation_eip1822.hex(), 16):
        as_proxy_for = _resolve_address(implementation_eip1822[-20:])
    elif implementation:
        # for other proxy patterns, we only check if etherscan indicates
        # the contract is a proxy. otherwise we could have a false positive
        # if there is an `implementation` method on a regular contract.
        try:
            # first try to call `implementation` per EIP897
            # https://eips.ethereum.org/EIPS/eip-897
            c = Contract.from_abi(name, address, abi)
            as_proxy_for = c.implementation.call()
        except Exception:
            # if that fails, fall back to the address provided by etherscan
            as_proxy_for = _resolve_address(implementation)

    if as_proxy_for:
        name, abi, _ = _extract_abi_data(as_proxy_for)
    return name, abi


def _setup_events(contract: Contract) -> None:
    """
    Helper function used to init contract event containers on a newly created `y.Contract` object.

    Args:
        contract: The contract object to set up events for.
    """
    if not hasattr(contract, "events"):
        contract.events = ContractEvents(contract)
    for k, v in contract.topics.items():
        setattr(contract.events, k, Events(addresses=[contract.address], topics=[[v]]))


_Address = NewType("_Address", str)
_unverified: Set[_Address] = set()
"""A collection of unverified addresses that is used to prevent repetitive etherscan api calls"""


_NOT_SYNCED = "`chain.height` returns 0 on your node, which means it is not fully synced.\nYou can only use this function on a fully synced node."



import asyncio
import logging
from typing import Any, Callable, KeysView, List, Optional, Tuple, Union

import a_sync
import brownie
from brownie.convert.datatypes import EthAddress

from y.contracts import Contract
from y.datatypes import Address, AddressOrContract, Block
from y.utils.multicall import multicall_decimals, multicall_totalSupply
from y.utils.raw_calls import _decimals, _totalSupply

logger = logging.getLogger(__name__)

SUPPORTED_INPUT_TYPES = str, Address, EthAddress, brownie.Contract, Contract


# These helpers can be used to fetch values for one or more tokens at once.
# NOTE You shoulnd't use these, they will likely be deleted soon for a cleaner alternative.


@a_sync.a_sync(default="sync")
async def decimals(
    contract_address_or_addresses: Union[
        AddressOrContract, List[AddressOrContract], Tuple[AddressOrContract]
    ],
    block: Optional[Block] = None,
    return_None_on_failure: bool = False,
):

    func = _choose_appropriate_fn(
        contract_address_or_addresses, _decimals, multicall_decimals
    )
    return await func(
        contract_address_or_addresses,
        block=block,
        return_None_on_failure=return_None_on_failure,
        sync=False,
    )


@a_sync.a_sync(default="sync")
async def totalSupply(
    contract_address_or_addresses: Union[
        AddressOrContract, List[AddressOrContract], Tuple[AddressOrContract]
    ],
    block: Optional[Block] = None,
    return_None_on_failure: bool = False,
):

    func = _choose_appropriate_fn(
        contract_address_or_addresses, _totalSupply, multicall_totalSupply
    )
    return await func(
        contract_address_or_addresses,
        block=block,
        return_None_on_failure=return_None_on_failure,
        sync=False,
    )


@a_sync.a_sync(default="sync")
async def totalSupplyReadable(
    contract_address_or_addresses: Union[
        AddressOrContract, List[AddressOrContract], Tuple[AddressOrContract]
    ],
    block: Optional[Block] = None,
    return_None_on_failure: bool = False,
):

    token_supplys, token_decimals = await asyncio.gather(
        totalSupply(
            contract_address_or_addresses,
            block=block,
            return_None_on_failure=return_None_on_failure,
            sync=False,
        ),
        decimals(
            contract_address_or_addresses,
            block=block,
            return_None_on_failure=return_None_on_failure,
            sync=False,
        ),
    )

    if type(token_supplys) == brownie.Wei:  # if only fetching totalSupply for one token
        supply = token_supplys
        decimal = token_decimals
        return supply / 10**decimal
    return [
        supply / 10**decimal for supply, decimal in zip(token_supplys, token_decimals)
    ]


def _choose_appropriate_fn(input: Any, singlecall_fn: Callable, multicall_fn: Callable):

    input_type = _input_type(input)
    if input_type == "single":
        return singlecall_fn
    elif input_type == "multi":
        return multicall_fn


def _input_type(input: Any) -> str:

    if type(input) in [list, tuple] or isinstance(input, KeysView):
        for value in input:
            _check_if_supported(value)
        return "multi"

    else:
        _check_if_supported(input)
        return "single"


def _check_if_supported(input: Any) -> bool:

    if type(input) not in SUPPORTED_INPUT_TYPES:
        raise TypeError(f"Unsupported input type: {type(input)}")



import asyncio
import logging
from typing import Optional, Tuple

import a_sync
from multicall import Call

from y import ENVIRONMENT_VARIABLES as ENVS
from y import convert
from y.classes.common import ERC20, WeiBalance
from y.contracts import has_methods
from y.datatypes import AnyAddressType, Block, UsdPrice, UsdValue
from y.exceptions import call_reverted
from y.utils import gather_methods
from y.utils.cache import optional_async_diskcache

_RESERVES_METHODS = (
    "token0()(address)",
    "token1()(address)",
    "usersAmounts()((uint,uint))",
)

logger = logging.getLogger(__name__)


@a_sync.a_sync(default="sync", cache_type="memory", ram_cache_ttl=5 * 60)
@optional_async_diskcache
async def is_popsicle_lp(token_address: AnyAddressType) -> bool:
    """
    Determines if the given token address is a Popsicle Finance LP token.

    Args:
        token_address: The address of the token to check.

    Returns:
        True if the token is a Popsicle LP token, False otherwise.

    Example:
        >>> is_popsicle = is_popsicle_lp("0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf")
        >>> print(is_popsicle)
        True
    """
    # NOTE: contract to check for reference (mainnet): 0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf
    return await has_methods(
        token_address,
        ("token0()(address)", "token1()(address)", "usersAmounts()((uint,uint))"),
        sync=False,
    )


@a_sync.a_sync(default="sync")
async def get_price(
    token: AnyAddressType,
    block: Optional[Block] = None,
    *,
    skip_cache: bool = ENVS.SKIP_CACHE
) -> Optional[UsdPrice]:
    """
    Calculates the price of a Popsicle Finance LP token.

    Args:
        token: The address of the Popsicle LP token.
        block (optional): The block number to query. Defaults to the latest block.
        skip_cache (optional): Whether to skip the cache when fetching prices. Defaults to :obj:`ENVS.SKIP_CACHE`.

    Returns:
        The price of the LP token in USD, or None if the price cannot be determined.

    Example:
        >>> price = get_price("0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf", block=14_000_000)
        >>> print(f"{price:.6f}")
        1.234567  # The price of the Popsicle LP token in USD
    """
    address = await convert.to_address_async(token)
    total_val = await get_tvl(address, block, skip_cache=skip_cache, sync=False)
    if total_val is None:
        return None
    total_supply = await ERC20(address, asynchronous=True).total_supply_readable(
        block, sync=False
    )
    return UsdPrice(total_val / total_supply)


@a_sync.a_sync(default="sync")
async def get_tvl(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
) -> Optional[UsdValue]:
    balances: Tuple[WeiBalance, WeiBalance]
    balances = await get_balances(
        token, block, skip_cache=skip_cache, _async_balance_objects=True, sync=False
    )
    return UsdValue(await WeiBalance.value_usd.sum(balances)) if balances else None


@a_sync.a_sync(default="sync")
async def get_balances(
    token: AnyAddressType,
    block: Optional[Block] = None,
    skip_cache: bool = ENVS.SKIP_CACHE,
    _async_balance_objects: bool = False,
) -> Optional[Tuple[WeiBalance, WeiBalance]]:
    try:
        token0, token1, (balance0, balance1) = await gather_methods(
            await convert.to_address_async(token), _RESERVES_METHODS, block=block
        )
    except Exception as e:
        if call_reverted(e):
            return None
        elif str(e) == "not enough values to unpack (expected 3, got 2)":
            # TODO determine if this is regular behavior when no tvl in pool or if this is bug to fix
            return None
        raise
    balance0 = WeiBalance(
        balance0,
        token0,
        block=block,
        skip_cache=skip_cache,
        asynchronous=_async_balance_objects,
    )
    balance1 = WeiBalance(
        balance1,
        token1,
        block=block,
        skip_cache=skip_cache,
        asynchronous=_async_balance_objects,
    )
    return balance0, balance1
